#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_serialization import jsonutils
from requests import exceptions

from heat.common import exception
from heat.common import grouputils
from heat.common.i18n import _
from heat.common import template_format
from heat.common import urlfetch
from heat.engine import attributes
from heat.engine import environment
from heat.engine import properties
from heat.engine.resources import stack_resource
from heat.engine import template
from heat.rpc import api as rpc_api

LOG = logging.getLogger(__name__)


REMOTE_SCHEMES = ('http', 'https')
LOCAL_SCHEMES = ('file',)

STACK_ID_OUTPUT = 'OS::stack_id'


def generate_class_from_template(name, data, param_defaults):
    tmpl = template.Template(template_format.parse(data))
    props, attrs = TemplateResource.get_schemas(tmpl, param_defaults)
    cls = type(name, (TemplateResource,),
               {'properties_schema': props,
                'attributes_schema': attrs,
                '__doc__': tmpl.t.get(tmpl.DESCRIPTION)})
    return cls


class TemplateResource(stack_resource.StackResource):
    """A resource implemented by a nested stack.

    This implementation passes resource properties as parameters to the nested
    stack. Outputs of the nested stack are exposed as attributes of this
    resource.
    """

    def __init__(self, name, json_snippet, stack):
        self._parsed_nested = None
        self.stack = stack
        self.validation_exception = None

        tri = self._get_resource_info(json_snippet)

        self.properties_schema = {}
        self.attributes_schema = {}

        # run Resource.__init__() so we can call self.nested()
        super(TemplateResource, self).__init__(name, json_snippet, stack)
        self.resource_info = tri
        if self.validation_exception is None:
            self._generate_schema()
            self.reparse()

    def _get_resource_info(self, rsrc_defn):
        try:
            tri = self.stack.env.get_resource_info(
                rsrc_defn.resource_type,
                resource_name=rsrc_defn.name,
                registry_type=environment.TemplateResourceInfo)
        except exception.EntityNotFound:
            self.validation_exception = ValueError(_(
                'Only Templates with an extension of .yaml or '
                '.template are supported'))
        else:
            self._template_name = tri.template_name
            self.resource_type = tri.name
            self.resource_path = tri.path
            if tri.user_resource:
                self.allowed_schemes = REMOTE_SCHEMES
            else:
                self.allowed_schemes = REMOTE_SCHEMES + LOCAL_SCHEMES

            return tri

    @staticmethod
    def get_template_file(template_name, allowed_schemes):
        try:
            return urlfetch.get(template_name, allowed_schemes=allowed_schemes)
        except (IOError, exceptions.RequestException) as r_exc:
            args = {'name': template_name, 'exc': str(r_exc)}
            msg = _('Could not fetch remote template '
                    '"%(name)s": %(exc)s') % args
            raise exception.NotFound(msg_fmt=msg)

    @staticmethod
    def get_schemas(tmpl, param_defaults):
        return ((properties.Properties.schema_from_params(
                tmpl.param_schemata(param_defaults))),
                (attributes.Attributes.schema_from_outputs(
                 tmpl[tmpl.OUTPUTS])))

    def _generate_schema(self):
        self._parsed_nested = None
        try:
            tmpl = template.Template(self.child_template())
        except (exception.NotFound, ValueError) as download_error:
            self.validation_exception = download_error
            tmpl = template.Template(
                {"HeatTemplateFormatVersion": "2012-12-12"})

        # re-generate the properties and attributes from the template.
        self.properties_schema, self.attributes_schema = self.get_schemas(
            tmpl, self.stack.env.param_defaults)

        self.attributes_schema.update(self.base_attributes_schema)
        self.attributes.set_schema(self.attributes_schema)

    def child_params(self):
        """Override method of child_params for the resource.

        :return: parameter values for our nested stack based on our properties
        """
        params = {}
        for pname, pval in iter(self.properties.props.items()):
            if not pval.implemented():
                continue

            try:
                val = self.properties.get_user_value(pname)
            except ValueError:
                if self.action == self.INIT:
                    prop = self.properties.props[pname]
                    val = prop.get_value(None)
                else:
                    raise

            if val is not None:
                # take a list and create a CommaDelimitedList
                if pval.type() == properties.Schema.LIST:
                    if len(val) == 0:
                        params[pname] = ''
                    elif isinstance(val[0], dict):
                        flattened = []
                        for (count, item) in enumerate(val):
                            for (ik, iv) in iter(item.items()):
                                mem_str = '.member.%d.%s=%s' % (count, ik, iv)
                                flattened.append(mem_str)
                        params[pname] = ','.join(flattened)
                    else:
                        # When None is returned from get_attr, creating a
                        # delimited list with it fails during validation.
                        # we should sanitize the None values to empty strings.
                        # FIXME(rabi) this needs a permanent solution
                        # to sanitize attributes and outputs in the future.
                        params[pname] = ','.join(
                            (x if x is not None else '') for x in val)
                else:
                    # for MAP, the JSON param takes either a collection or
                    # string, so just pass it on and let the param validate
                    # as appropriate
                    params[pname] = val

        return params

    def child_template(self):
        if not self._parsed_nested:
            self._parsed_nested = template_format.parse(self.template_data(),
                                                        self.template_url)
        return self._parsed_nested

    def regenerate_info_schema(self, definition):
        self._get_resource_info(definition)
        self._generate_schema()

    @property
    def template_url(self):
        return self._template_name

    def template_data(self):
        # we want to have the latest possible template.
        # 1. look in files
        # 2. try download
        # 3. look in the db
        reported_excp = None
        t_data = self.stack.t.files.get(self.template_url)
        stored_t_data = t_data

        if t_data is None:
            LOG.debug('TemplateResource data file "%s" not found in files.',
                      self.template_url)
        if not t_data and self.template_url.endswith((".yaml", ".template")):
            try:
                t_data = self.get_template_file(self.template_url,
                                                self.allowed_schemes)
            except exception.NotFound as err:
                if self.action == self.UPDATE:
                    raise
                reported_excp = err

        if t_data is None:
            nested_identifier = self.nested_identifier()
            if nested_identifier is not None:
                nested_t = self.rpc_client().get_template(self.context,
                                                          nested_identifier)
                t_data = jsonutils.dumps(nested_t)

        if t_data is not None:
            if t_data != stored_t_data:
                self.stack.t.files[self.template_url] = t_data
            self.stack.t.env.register_class(self.resource_type,
                                            self.template_url,
                                            path=self.resource_path)
            return t_data
        if reported_excp is None:
            reported_excp = ValueError(_('Unknown error retrieving %s') %
                                       self.template_url)
        raise reported_excp

    def _validate_against_facade(self, facade_cls):
        facade_schemata = properties.schemata(facade_cls.properties_schema)

        for n, fs in facade_schemata.items():
            if fs.required and n not in self.properties_schema:
                msg = (_("Required property %(n)s for facade %(type)s "
                       "missing in provider") % {'n': n, 'type': self.type()})
                raise exception.StackValidationFailed(message=msg)

            ps = self.properties_schema.get(n)
            if (n in self.properties_schema and
                    (fs.allowed_param_prop_type() != ps.type)):
                # Type mismatch
                msg = (_("Property %(n)s type mismatch between facade %(type)s"
                       " (%(fs_type)s) and provider (%(ps_type)s)") % {
                           'n': n, 'type': self.type(),
                           'fs_type': fs.type, 'ps_type': ps.type})
                raise exception.StackValidationFailed(message=msg)

        for n, ps in self.properties_schema.items():
            if ps.required and n not in facade_schemata:
                # Required property for template not present in facade
                msg = (_("Provider requires property %(n)s "
                       "unknown in facade %(type)s") % {
                           'n': n, 'type': self.type()})
                raise exception.StackValidationFailed(message=msg)

        facade_attrs = facade_cls.attributes_schema.copy()
        facade_attrs.update(facade_cls.base_attributes_schema)
        for attr in facade_attrs:
            if attr not in self.attributes_schema:
                msg = (_("Attribute %(attr)s for facade %(type)s "
                       "missing in provider") % {
                           'attr': attr, 'type': self.type()})
                raise exception.StackValidationFailed(message=msg)

    def validate(self):
        # Calls validate_template()
        result = super(TemplateResource, self).validate()

        try:
            self.template_data()
        except ValueError as ex:
            msg = _("Failed to retrieve template data: %s") % ex
            raise exception.StackValidationFailed(message=msg)

        # If we're using an existing resource type as a facade for this
        # template, check for compatibility between the interfaces.
        try:
            fri = self.stack.env.get_resource_info(
                self.type(),
                resource_name=self.name,
                ignore=self.resource_info)
        except exception.EntityNotFound:
            pass
        else:
            facade_cls = fri.get_class(files=self.stack.t.files)
            self._validate_against_facade(facade_cls)

        return result

    def validate_template(self):
        if self.validation_exception is not None:
            msg = str(self.validation_exception)
            raise exception.StackValidationFailed(message=msg)

        return super(TemplateResource, self).validate_template()

    def handle_adopt(self, resource_data=None):
        return self.create_with_template(self.child_template(),
                                         self.child_params(),
                                         adopt_data=resource_data)

    def handle_create(self):
        return self.create_with_template(self.child_template(),
                                         self.child_params())

    def metadata_update(self, new_metadata=None):
        """Refresh the metadata if new_metadata is None."""
        if new_metadata is None:
            self.metadata_set(self.t.metadata())

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        self.properties = json_snippet.properties(self.properties_schema,
                                                  self.context)
        return self.update_with_template(self.child_template(),
                                         self.child_params())

    def get_reference_id(self):
        if self.resource_id is None:
            return str(self.name)

        if STACK_ID_OUTPUT in self.attributes.cached_attrs:
            return self.attributes.cached_attrs[STACK_ID_OUTPUT]

        stack_identity = self.nested_identifier()
        reference_id = stack_identity.arn()

        try:
            if self._outputs is not None:
                reference_id = self.get_output(STACK_ID_OUTPUT)
            elif STACK_ID_OUTPUT in self.attributes:
                output = self.rpc_client().show_output(self.context,
                                                       dict(stack_identity),
                                                       STACK_ID_OUTPUT)
                if rpc_api.OUTPUT_ERROR in output:
                    raise exception.TemplateOutputError(
                        resource=self.name,
                        attribute=STACK_ID_OUTPUT,
                        message=output[rpc_api.OUTPUT_ERROR])
                reference_id = output[rpc_api.OUTPUT_VALUE]
        except exception.TemplateOutputError as err:
            LOG.info('%s', err)
        except exception.NotFound:
            pass

        self.attributes.set_cached_attr(STACK_ID_OUTPUT, reference_id)
        return reference_id

    def get_attribute(self, key, *path):
        if self.resource_id is None:
            return None

        # first look for explicit resource.x.y
        if key.startswith('resource.'):
            return grouputils.get_nested_attrs(self, key, False, *path)

        # then look for normal outputs
        try:
            return attributes.select_from_attribute(self.get_output(key),
                                                    path)
        except exception.NotFound:
            raise exception.InvalidTemplateAttribute(resource=self.name,
                                                     key=key)




#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutronclient.v2_0 import client as neutronclient

from heat.common import template_format
from heat.engine.clients.os import neutron
from heat.engine.resources.openstack.neutron import extraroute
from heat.engine import scheduler
from heat.tests import common
from heat.tests import utils


neutron_template = '''
{
  "AWSTemplateFormatVersion" : "2010-09-09",
  "Description" : "Template to test OS::Neutron::ExtraRoute resources",
  "Parameters" : {},
  "Resources" : {
    "router": {
      "Type": "OS::Neutron::Router"
    },
    "extraroute1": {
      "Type": "OS::Neutron::ExtraRoute",
      "Properties": {
        "router_id": { "Ref" : "router" },
        "destination" : "192.168.0.0/24",
        "nexthop": "1.1.1.1"
      }
    },
    "extraroute2": {
      "Type": "OS::Neutron::ExtraRoute",
      "Properties": {
        "router_id": { "Ref" : "router" },
        "destination" : "192.168.255.0/24",
        "nexthop": "1.1.1.1"
      }
    }
  }
}
'''


class NeutronExtraRouteTest(common.HeatTestCase):
    def setUp(self):
        super(NeutronExtraRouteTest, self).setUp()
        self.mockclient = mock.Mock(spec=neutronclient.Client)
        self.patchobject(neutronclient, 'Client',
                         return_value=self.mockclient)

        self.patchobject(neutron.NeutronClientPlugin, 'has_extension',
                         return_value=True)

    def create_extraroute(self, t, stack, resource_name, properties=None):
        properties = properties or {}
        t['Resources'][resource_name]['Properties'] = properties
        rsrc = extraroute.ExtraRoute(
            resource_name,
            stack.t.resource_definitions(stack)[resource_name],
            stack)
        scheduler.TaskRunner(rsrc.create)()
        self.assertEqual((rsrc.CREATE, rsrc.COMPLETE), rsrc.state)
        return rsrc

    def _test_extraroute(self, ipv6=False):

        if ipv6:
            route1 = {"destination": "ffff:f53b:82e4::56/46",
                      "nexthop": "dce7:f53b:82e4::56"}
            route2 = {"destination": "ffff:f53b:ffff::56/46",
                      "nexthop": "dce7:f53b:82e4::56"}
        else:
            route1 = {"destination": "192.168.0.0/24",
                      "nexthop": "1.1.1.1"}
            route2 = {"destination": "192.168.255.0/24",
                      "nexthop": "1.1.1.1"}

        self.stub_RouterConstraint_validate()

        self.mockclient.show_router.side_effect = [
            # add first route
            {'router': {'routes': []}},
            # add second route
            {'router': {'routes': [route1.copy()]}},
            # first delete
            {'router': {'routes': [route1.copy(), route2.copy()]}},
            # second delete
            {'router': {'routes': [route2.copy()]}},
        ]
        self.mockclient.update_router.return_value = None

        t = template_format.parse(neutron_template)
        stack = utils.parse_stack(t)

        if ipv6:
            rsrc1 = self.create_extraroute(
                t, stack, 'extraroute1', properties={
                    'router_id': '3e46229d-8fce-4733-819a-b5fe630550f8',
                    'destination': 'ffff:f53b:82e4::56/46',
                    'nexthop': 'dce7:f53b:82e4::56'})

            self.create_extraroute(
                t, stack, 'extraroute2', properties={
                    'router_id': '3e46229d-8fce-4733-819a-b5fe630550f8',
                    'destination': 'ffff:f53b:ffff::56/46',
                    'nexthop': 'dce7:f53b:82e4::56'})
        else:
            rsrc1 = self.create_extraroute(
                t, stack, 'extraroute1', properties={
                    'router_id': '3e46229d-8fce-4733-819a-b5fe630550f8',
                    'destination': '192.168.0.0/24',
                    'nexthop': '1.1.1.1'})

            self.create_extraroute(
                t, stack, 'extraroute2', properties={
                    'router_id': '3e46229d-8fce-4733-819a-b5fe630550f8',
                    'destination': '192.168.255.0/24',
                    'nexthop': '1.1.1.1'})

        scheduler.TaskRunner(rsrc1.delete)()
        rsrc1.state_set(rsrc1.CREATE, rsrc1.COMPLETE, 'to delete again')
        scheduler.TaskRunner(rsrc1.delete)()

        self.mockclient.show_router.assert_called_with(
            '3e46229d-8fce-4733-819a-b5fe630550f8')
        self.mockclient.update_router.assert_has_calls([
            # add first route
            mock.call('3e46229d-8fce-4733-819a-b5fe630550f8',
                      {'router': {'routes': [route1.copy()]}}),
            # add second route
            mock.call('3e46229d-8fce-4733-819a-b5fe630550f8',
                      {'router': {'routes': [route1.copy(),
                                             route2.copy()]}}),
            # first delete
            mock.call('3e46229d-8fce-4733-819a-b5fe630550f8',
                      {'router': {'routes': [route2.copy()]}}),
        ])

    def test_extraroute_ipv4(self):
        self._test_extraroute(ipv6=False)

    def test_extraroute_ipv6(self):
        self._test_extraroute(ipv6=True)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from heat.common import template_format
from heat.engine.clients.os import neutron
from heat.engine import rsrc_defn
from heat.engine import stack
from heat.engine import template
from heat.tests import common
from heat.tests import utils

address_scope_template = '''
heat_template_version: 2016-04-08
description: This template to define a neutron address scope.
resources:
  my_address_scope:
    type: OS::Neutron::AddressScope
    properties:
      shared: False
      tenant_id: d66c74c01d6c41b9846088c1ad9634d0
'''


class NeutronAddressScopeTest(common.HeatTestCase):
    def setUp(self):
        super(NeutronAddressScopeTest, self).setUp()

        self.ctx = utils.dummy_context()
        tpl = template_format.parse(address_scope_template)
        self.stack = stack.Stack(
            self.ctx,
            'neutron_address_scope_test',
            template.Template(tpl)
        )

        self.neutronclient = mock.MagicMock()
        self.patchobject(neutron.NeutronClientPlugin, 'has_extension',
                         return_value=True)
        self.my_address_scope = self.stack['my_address_scope']
        self.my_address_scope.client = mock.MagicMock(
            return_value=self.neutronclient)
        self.patchobject(self.my_address_scope, 'physical_resource_name',
                         return_value='test_address_scope')

    def test_address_scope_handle_create(self):
        addrs = {
            'address_scope': {
                'id': '9c1eb3fe-7bba-479d-bd43-1d497e53c384',
                'tenant_id': 'd66c74c01d6c41b9846088c1ad9634d0',
                'shared': False,
                'ip_version': 4
            }
        }
        create_props = {'name': 'test_address_scope',
                        'shared': False,
                        'tenant_id': 'd66c74c01d6c41b9846088c1ad9634d0',
                        'ip_version': 4}

        self.neutronclient.create_address_scope.return_value = addrs
        self.my_address_scope.handle_create()
        self.assertEqual('9c1eb3fe-7bba-479d-bd43-1d497e53c384',
                         self.my_address_scope.resource_id)
        self.neutronclient.create_address_scope.assert_called_once_with(
            {'address_scope': create_props}
        )

    def test_address_scope_handle_delete(self):
        addrs_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'
        self.my_address_scope.resource_id = addrs_id
        self.neutronclient.delete_address_scope.return_value = None

        self.assertIsNone(self.my_address_scope.handle_delete())
        self.neutronclient.delete_address_scope.assert_called_once_with(
            self.my_address_scope.resource_id)

    def test_address_scope_handle_delete_not_found(self):
        addrs_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'
        self.my_address_scope.resource_id = addrs_id
        not_found = self.neutronclient.NotFound
        self.neutronclient.delete_address_scope.side_effect = not_found

        self.assertIsNone(self.my_address_scope.handle_delete())
        self.neutronclient.delete_address_scope.assert_called_once_with(
            self.my_address_scope.resource_id)

    def test_address_scope_handle_delete_resource_id_is_none(self):
        self.my_address_scope.resource_id = None
        self.assertIsNone(self.my_address_scope.handle_delete())
        self.assertEqual(0,
                         self.neutronclient.delete_address_scope.call_count)

    def test_address_scope_handle_update(self):
        addrs_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'
        self.my_address_scope.resource_id = addrs_id

        props = {
            'name': 'test_address_scope',
            'shared': True
        }

        update_dict = props.copy()
        update_snippet = rsrc_defn.ResourceDefinition(
            self.my_address_scope.name,
            self.my_address_scope.type(),
            props)

        # with name
        self.my_address_scope.handle_update(
            json_snippet=update_snippet,
            tmpl_diff={},
            prop_diff=props)

        # without name
        props['name'] = None
        self.my_address_scope.handle_update(
            json_snippet=update_snippet,
            tmpl_diff={},
            prop_diff=props)
        self.assertEqual(2, self.neutronclient.update_address_scope.call_count)
        self.neutronclient.update_address_scope.assert_called_with(
            addrs_id, {'address_scope': update_dict})

    def test_address_scope_get_attr(self):
        self.my_address_scope.resource_id = 'addrs_id'
        addrs = {
            'address_scope': {
                'name': 'test_addrs',
                'id': '9c1eb3fe-7bba-479d-bd43-1d497e53c384',
                'tenant_id': 'd66c74c01d6c41b9846088c1ad9634d0',
                'shared': True,
                'ip_version': 4
            }
        }
        self.neutronclient.show_address_scope.return_value = addrs
        self.assertEqual(addrs['address_scope'],
                         self.my_address_scope.FnGetAtt('show'))
        self.neutronclient.show_address_scope.assert_called_once_with(
            self.my_address_scope.resource_id)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common.i18n import _
from heat.engine import constraints
from heat.engine import properties
from heat.engine.resources import alarm_base
from heat.engine import support


class CompositeAlarm(alarm_base.BaseAlarm):
    """A resource that implements Aodh composite alarm.

    Allows to specify multiple rules when creating a composite alarm,
    and the rules combined with logical operators: and, or.
    """

    alarm_type = 'composite'

    support_status = support.SupportStatus(version='8.0.0')

    PROPERTIES = (
        COMPOSITE_RULE, COMPOSITE_OPERATOR, RULES,
    ) = (
        'composite_rule', 'operator', 'rules',
    )

    composite_rule_schema = {
        COMPOSITE_OPERATOR: properties.Schema(
            properties.Schema.STRING,
            _('The operator indicates how to combine the rules.'),
            required=True,
            update_allowed=True,
            constraints=[
                constraints.AllowedValues(['or', 'and'])
            ]
        ),
        RULES: properties.Schema(
            properties.Schema.LIST,
            _('Rules list. Basic threshold/gnocchi rules and nested dict '
              'which combine threshold/gnocchi rules by "and" or "or" are '
              'allowed. For example, the form is like: [RULE1, RULE2, '
              '{"and": [RULE3, RULE4]}], the basic threshold/gnocchi '
              'rules must include a "type" field.'),
            required=True,
            update_allowed=True,
            constraints=[
                constraints.Length(min=2)
            ]
        ),
    }

    properties_schema = {
        COMPOSITE_RULE: properties.Schema(
            properties.Schema.MAP,
            _('Composite threshold rules in JSON format.'),
            required=True,
            update_allowed=True,
            schema=composite_rule_schema
        )
    }

    properties_schema.update(alarm_base.common_properties_schema)

    def parse_composite_rule(self, props):
        composite_rule = props.get(self.COMPOSITE_RULE)
        operator = composite_rule[self.COMPOSITE_OPERATOR]
        rules = composite_rule[self.RULES]
        props[self.COMPOSITE_RULE] = {operator: rules}

    def handle_create(self):
        props = self.actions_to_urls(self.properties)
        self.parse_composite_rule(props)
        props['name'] = self.physical_resource_name()
        props['type'] = self.alarm_type

        alarm = self.client().alarm.create(props)
        self.resource_id_set(alarm['alarm_id'])

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        if prop_diff:
            updated_props = self.actions_to_urls(prop_diff)
            if self.COMPOSITE_RULE in prop_diff:
                self.parse_composite_rule(updated_props)
            self.client().alarm.update(self.resource_id, updated_props)


def resource_mapping():
    return {
        'OS::Aodh::CompositeAlarm': CompositeAlarm,
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import uuid

from heat.common import template_format
from heat.engine.clients.os import glance
from heat.engine.clients.os import neutron
from heat.engine.clients.os import nova
from heat.engine import environment
from heat.engine.resources.aws.ec2 import instance as instances
from heat.engine.resources.aws.ec2 import network_interface as net_interfaces
from heat.engine import scheduler
from heat.engine import stack as parser
from heat.engine import template
from heat.tests import common
from heat.tests.openstack.nova import fakes as fakes_nova
from heat.tests import utils

wp_template = '''
{
  "AWSTemplateFormatVersion" : "2010-09-09",
  "Description" : "WordPress",
  "Parameters" : {
    "KeyName" : {
      "Description" : "KeyName",
      "Type" : "String",
      "Default" : "test"
    },
    "InstanceType": {
      "Type": "String",
      "Description": "EC2 instance type",
      "Default": "m1.small",
      "AllowedValues": [ "m1.small", "m1.large" ]
    },
    "SubnetId": {
      "Type" : "String",
      "Description" : "SubnetId of an existing subnet in your VPC"
    },
  },
  "Resources" : {
    "WebServer": {
      "Type": "AWS::EC2::Instance",
      "Properties": {
        "ImageId"        : "F17-x86_64-gold",
        "InstanceType"   : { "Ref" : "InstanceType" },
        "SubnetId"       : { "Ref" : "SubnetId" },
        "KeyName"        : { "Ref" : "KeyName" },
        "UserData"       : "wordpress"
      }
    }
  }
}
'''


wp_template_with_nic = '''
{
  "AWSTemplateFormatVersion" : "2010-09-09",
  "Description" : "WordPress",
  "Parameters" : {
    "KeyName" : {
      "Description" : "KeyName",
      "Type" : "String",
      "Default" : "test"
    },
    "InstanceType": {
      "Type": "String",
      "Description": "EC2 instance type",
      "Default": "m1.small",
      "AllowedValues": [ "m1.small", "m1.large" ]
    },
    "SubnetId": {
      "Type" : "String",
      "Description" : "SubnetId of an existing subnet in your VPC"
    },
  },
  "Resources" : {

    "nic1": {
        "Type": "AWS::EC2::NetworkInterface",
        "Properties": {
            "SubnetId": { "Ref": "SubnetId" }
        }
    },

    "WebServer": {
      "Type": "AWS::EC2::Instance",
      "Properties": {
        "ImageId"        : "F17-x86_64-gold",
        "InstanceType"   : { "Ref" : "InstanceType" },
        "NetworkInterfaces": [ { "NetworkInterfaceId" : {"Ref": "nic1"},
                                 "DeviceIndex" : "0"  } ],
        "KeyName"        : { "Ref" : "KeyName" },
        "UserData"       : "wordpress"
      }
    }
  }
}
'''


class FakeNeutron(object):

    def show_subnet(self, subnet, **_params):
        return {
            'subnet': {
                'name': 'name',
                'network_id': 'fc68ea2c-b60b-4b4f-bd82-94ec81110766',
                'tenant_id': 'c1210485b2424d48804aad5d39c61b8f',
                'allocation_pools': [{'start': '10.10.0.2',
                                      'end': '10.10.0.254'}],
                'gateway_ip': '10.10.0.1',
                'ip_version': 4,
                'cidr': '10.10.0.0/24',
                'id': '4156c7a5-e8c4-4aff-a6e1-8f3c7bc83861',
                'enable_dhcp': False,
            }}

    def create_port(self, body=None):
        return {
            'port': {
                'admin_state_up': True,
                'device_id': '',
                'device_owner': '',
                'fixed_ips': [{
                    'ip_address': '10.0.3.3',
                    'subnet_id': '4156c7a5-e8c4-4aff-a6e1-8f3c7bc83861'}],
                'id': '64d913c1-bcb1-42d2-8f0a-9593dbcaf251',
                'mac_address': 'fa:16:3e:25:32:5d',
                'name': '',
                'network_id': 'fc68ea2c-b60b-4b4f-bd82-94ec81110766',
                'status': 'ACTIVE',
                'tenant_id': 'c1210485b2424d48804aad5d39c61b8f'
            }}

    def delete_port(self, port_id):
        return None


class instancesTest(common.HeatTestCase):
    def setUp(self):
        super(instancesTest, self).setUp()
        self.fc = fakes_nova.FakeClient()

    def _mock_get_image_id_success(self, imageId_input, imageId):
        self.m_f_i = self.patchobject(glance.GlanceClientPlugin,
                                      'find_image_by_name_or_id',
                                      return_value=imageId)

    def _test_instance_create_delete(self, vm_status='ACTIVE',
                                     vm_delete_status='NotFound'):
        return_server = self.fc.servers.list()[1]
        instance = self._create_test_instance(return_server, 'in_create')

        instance.resource_id = '1234'
        instance.status = vm_status
        # this makes sure the auto increment worked on instance creation
        self.assertGreater(instance.id, 0)

        expected_ip = return_server.networks['public'][0]
        self.assertEqual(expected_ip, instance.FnGetAtt('PublicIp'))
        self.assertEqual(expected_ip, instance.FnGetAtt('PrivateIp'))
        self.assertEqual(expected_ip, instance.FnGetAtt('PrivateDnsName'))
        self.assertEqual(expected_ip, instance.FnGetAtt('PublicDnsName'))

        d1 = {'server': self.fc.client.get_servers_detail()[1]['servers'][0]}
        d1['server']['status'] = vm_status

        m_gs_side_effects = [(200, d1)]

        d2 = copy.deepcopy(d1)
        if vm_delete_status == 'DELETED':
            d2['server']['status'] = vm_delete_status
            m_gs_side_effects.append((200, d2))
        else:
            m_gs_side_effects.append(fakes_nova.fake_exception)

        self.patchobject(self.fc.client, 'get_servers_1234',
                         side_effect=m_gs_side_effects)
        scheduler.TaskRunner(instance.delete)()
        self.assertEqual((instance.DELETE, instance.COMPLETE), instance.state)
        self.assertEqual(2, self.fc.client.get_servers_1234.call_count)

    def _create_test_instance(self, return_server, name):
        stack_name = '%s_s' % name
        t = template_format.parse(wp_template)
        kwargs = {'KeyName': 'test',
                  'InstanceType': 'm1.large',
                  'SubnetId': '4156c7a5-e8c4-4aff-a6e1-8f3c7bc83861'}
        tmpl = template.Template(t,
                                 env=environment.Environment(kwargs))
        self.stack = parser.Stack(utils.dummy_context(), stack_name, tmpl,
                                  stack_id=str(uuid.uuid4()))
        image_id = 'CentOS 5.2'
        t['Resources']['WebServer']['Properties']['ImageId'] = image_id
        resource_defns = self.stack.t.resource_definitions(self.stack)
        instance = instances.Instance('%s_name' % name,
                                      resource_defns['WebServer'], self.stack)
        metadata = instance.metadata_get()

        self.patchobject(nova.NovaClientPlugin, 'client',
                         return_value=self.fc)

        self._mock_get_image_id_success(image_id, 1)
        self.stub_SubnetConstraint_validate()
        self.patchobject(instance, 'neutron', return_value=FakeNeutron())

        self.patchobject(neutron.NeutronClientPlugin, '_create',
                         return_value=FakeNeutron())

        # need to resolve the template functions
        server_userdata = instance.client_plugin().build_userdata(
            metadata,
            instance.properties['UserData'],
            'ec2-user')
        self.patchobject(nova.NovaClientPlugin, 'build_userdata',
                         return_value=server_userdata)
        self.patchobject(self.fc.servers, 'create', return_value=return_server)

        scheduler.TaskRunner(instance.create)()
        self.m_f_i.assert_called_with(image_id)
        self.fc.servers.create.assert_called_once_with(
            image=1, flavor=3, key_name='test',
            name=utils.PhysName(stack_name, instance.name),
            security_groups=None,
            userdata=server_userdata, scheduler_hints=None, meta=None,
            nics=[{'port-id': '64d913c1-bcb1-42d2-8f0a-9593dbcaf251'}],
            availability_zone=None,
            block_device_mapping=None)
        nova.NovaClientPlugin.build_userdata.assert_called_once_with(
            metadata,
            instance.properties['UserData'],
            'ec2-user')
        neutron.NeutronClientPlugin._create.assert_called_once_with()
        nova.NovaClientPlugin.client.assert_called_with()
        glance.GlanceClientPlugin.find_image_by_name_or_id.assert_called_with(
            image_id)
        return instance

    def _create_test_instance_with_nic(self, return_server, name):
        stack_name = '%s_s' % name
        t = template_format.parse(wp_template_with_nic)
        kwargs = {'KeyName': 'test',
                  'InstanceType': 'm1.large',
                  'SubnetId': '4156c7a5-e8c4-4aff-a6e1-8f3c7bc83861'}
        tmpl = template.Template(t,
                                 env=environment.Environment(kwargs))
        self.stack = parser.Stack(utils.dummy_context(), stack_name, tmpl,
                                  stack_id=str(uuid.uuid4()))
        image_id = 'CentOS 5.2'
        t['Resources']['WebServer']['Properties']['ImageId'] = image_id

        resource_defns = self.stack.t.resource_definitions(self.stack)
        nic = net_interfaces.NetworkInterface('%s_nic' % name,
                                              resource_defns['nic1'],
                                              self.stack)

        instance = instances.Instance('%s_name' % name,
                                      resource_defns['WebServer'], self.stack)
        metadata = instance.metadata_get()

        self._mock_get_image_id_success(image_id, 1)
        self.stub_SubnetConstraint_validate()
        self.patchobject(nic, 'client', return_value=FakeNeutron())

        self.patchobject(neutron.NeutronClientPlugin, '_create',
                         return_value=FakeNeutron())

        self.patchobject(nova.NovaClientPlugin, 'client',
                         return_value=self.fc)

        # need to resolve the template functions
        server_userdata = instance.client_plugin().build_userdata(
            metadata,
            instance.properties['UserData'],
            'ec2-user')
        self.patchobject(nova.NovaClientPlugin, 'build_userdata',
                         return_value=server_userdata)
        self.patchobject(self.fc.servers, 'create', return_value=return_server)

        # create network interface
        scheduler.TaskRunner(nic.create)()
        self.stack.resources["nic1"] = nic

        scheduler.TaskRunner(instance.create)()

        self.fc.servers.create.assert_called_once_with(
            image=1, flavor=3, key_name='test',
            name=utils.PhysName(stack_name, instance.name),
            security_groups=None,
            userdata=server_userdata, scheduler_hints=None, meta=None,
            nics=[{'port-id': '64d913c1-bcb1-42d2-8f0a-9593dbcaf251'}],
            availability_zone=None,
            block_device_mapping=None)
        self.m_f_i.assert_called_with(image_id)
        nova.NovaClientPlugin.build_userdata.assert_called_once_with(
            metadata,
            instance.properties['UserData'],
            'ec2-user')
        neutron.NeutronClientPlugin._create.assert_called_once_with()
        nova.NovaClientPlugin.client.assert_called_with()
        glance.GlanceClientPlugin.find_image_by_name_or_id.assert_called_with(
            image_id)
        return instance

    def test_instance_create_delete_with_SubnetId(self):
        self._test_instance_create_delete(vm_delete_status='DELETED')

    def test_instance_create_with_nic(self):
        return_server = self.fc.servers.list()[1]
        instance = self._create_test_instance_with_nic(
            return_server, 'in_create_wnic')

        # this makes sure the auto increment worked on instance creation
        self.assertGreater(instance.id, 0)

        expected_ip = return_server.networks['public'][0]
        self.assertEqual(expected_ip, instance.FnGetAtt('PublicIp'))
        self.assertEqual(expected_ip, instance.FnGetAtt('PrivateIp'))
        self.assertEqual(expected_ip, instance.FnGetAtt('PrivateDnsName'))
        self.assertEqual(expected_ip, instance.FnGetAtt('PublicDnsName'))



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from webob import exc

from heat.api.openstack.v1 import util
from heat.common import serializers
from heat.common import wsgi
from heat.rpc import client as rpc_client


class SoftwareDeploymentController(object):
    """WSGI controller for Software deployments in Heat v1 API.

    Implements the API actions.
    """

    REQUEST_SCOPE = 'software_deployments'

    def __init__(self, options):
        self.options = options
        self.rpc_client = rpc_client.EngineClient()

    def default(self, req, **args):
        raise exc.HTTPNotFound()

    @util.registered_policy_enforce
    def index(self, req):
        """List software deployments."""
        param_types = {
            'server_id': util.PARAM_TYPE_SINGLE,
        }
        params = util.get_allowed_params(req.params, param_types)
        sds = self.rpc_client.list_software_deployments(req.context, **params)
        return {'software_deployments': sds}

    @util.registered_policy_enforce
    def metadata(self, req, server_id):
        """List software deployments grouped by the group name.

        This is done for the requested server.
        """
        sds = self.rpc_client.metadata_software_deployments(
            req.context, server_id=server_id)
        return {'metadata': sds}

    @util.registered_policy_enforce
    def show(self, req, deployment_id):
        """Gets detailed information for a software deployment."""
        sd = self.rpc_client.show_software_deployment(req.context,
                                                      deployment_id)
        return {'software_deployment': sd}

    @util.registered_policy_enforce
    def create(self, req, body):
        """Create a new software deployment."""
        create_data = dict((k, body.get(k)) for k in (
            'config_id', 'server_id', 'input_values',
            'action', 'status', 'status_reason', 'stack_user_project_id'))

        sd = self.rpc_client.create_software_deployment(req.context,
                                                        **create_data)
        return {'software_deployment': sd}

    @util.registered_policy_enforce
    def update(self, req, deployment_id, body):
        """Update an existing software deployment."""
        update_data = dict((k, body.get(k)) for k in (
            'config_id', 'input_values', 'output_values', 'action',
            'status', 'status_reason')
            if body.get(k, None) is not None)
        sd = self.rpc_client.update_software_deployment(req.context,
                                                        deployment_id,
                                                        **update_data)
        return {'software_deployment': sd}

    @util.registered_policy_enforce
    def delete(self, req, deployment_id):
        """Delete an existing software deployment."""
        res = self.rpc_client.delete_software_deployment(req.context,
                                                         deployment_id)

        if res is not None:
            raise exc.HTTPBadRequest(res['Error'])

        raise exc.HTTPNoContent()


def create_resource(options):
    """Software deployments resource factory method."""
    deserializer = wsgi.JSONRequestDeserializer()
    serializer = serializers.JSONResponseSerializer()
    return wsgi.Resource(
        SoftwareDeploymentController(options), deserializer, serializer)



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import eventlet.queue
import functools

from oslo_log import log as logging
import oslo_messaging
from oslo_utils import excutils
from oslo_utils import uuidutils
from osprofiler import profiler

from heat.common import context
from heat.common import messaging as rpc_messaging
from heat.db import api as db_api
from heat.engine import check_resource
from heat.engine import node_data
from heat.engine import stack as parser
from heat.engine import sync_point
from heat.objects import stack as stack_objects
from heat.rpc import api as rpc_api
from heat.rpc import worker_client as rpc_client

LOG = logging.getLogger(__name__)

CANCEL_RETRIES = 3


def log_exceptions(func):
    @functools.wraps(func)
    def exception_wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception('Unhandled exception in %s', func.__name__)

    return exception_wrapper


@profiler.trace_cls("rpc")
class WorkerService(object):
    """Service that has 'worker' actor in convergence.

    This service is dedicated to handle internal messages to the 'worker'
    (a.k.a. 'converger') actor in convergence. Messages on this bus will
    use the 'cast' rather than 'call' method to anycast the message to
    an engine that will handle it asynchronously. It won't wait for
    or expect replies from these messages.
    """

    RPC_API_VERSION = '1.4'

    def __init__(self,
                 host,
                 topic,
                 engine_id,
                 thread_group_mgr):
        self.host = host
        self.topic = topic
        self.engine_id = engine_id
        self.thread_group_mgr = thread_group_mgr

        self._rpc_client = rpc_client.WorkerClient()
        self._rpc_server = None
        self.target = None

    def start(self):
        target = oslo_messaging.Target(
            version=self.RPC_API_VERSION,
            server=self.engine_id,
            topic=self.topic)
        self.target = target
        LOG.info("Starting %(topic)s (%(version)s) in engine %(engine)s.",
                 {'topic': self.topic,
                  'version': self.RPC_API_VERSION,
                  'engine': self.engine_id})

        self._rpc_server = rpc_messaging.get_rpc_server(target, self)
        self._rpc_server.start()

    def stop(self):
        if self._rpc_server is None:
            return
        # Stop rpc connection at first for preventing new requests
        LOG.info("Stopping %(topic)s in engine %(engine)s.",
                 {'topic': self.topic, 'engine': self.engine_id})
        try:
            self._rpc_server.stop()
            self._rpc_server.wait()
        except Exception as e:
            LOG.error("%(topic)s is failed to stop, %(exc)s",
                      {'topic': self.topic, 'exc': e})

    def stop_traversal(self, stack):
        """Update current traversal to stop workers from propagating.

        Marks the stack as FAILED due to cancellation, but, allows all
        in_progress resources to complete normally; no worker is stopped
        abruptly.

        Any in-progress traversals are also stopped on all nested stacks that
        are descendants of the one passed.
        """
        _stop_traversal(stack)

        db_child_stacks = stack_objects.Stack.get_all_by_root_owner_id(
            stack.context, stack.id)

        for db_child in db_child_stacks:
            if db_child.status == parser.Stack.IN_PROGRESS:
                child = parser.Stack.load(stack.context,
                                          stack_id=db_child.id,
                                          stack=db_child,
                                          load_template=False)
                _stop_traversal(child)

    def stop_all_workers(self, stack):
        """Cancel all existing worker threads for the stack.

        Threads will stop running at their next yield point, whether or not the
        resource operations are complete.
        """
        cancelled = _cancel_workers(stack, self.thread_group_mgr,
                                    self.engine_id, self._rpc_client)
        if not cancelled:
            LOG.error("Failed to stop all workers of stack %s, "
                      "stack cancel not complete", stack.name)
            return False

        LOG.info('[%(name)s(%(id)s)] Stopped all active workers for stack '
                 '%(action)s',
                 {'name': stack.name, 'id': stack.id, 'action': stack.action})

        return True

    def _retrigger_replaced(self, is_update, rsrc, stack, check_resource):
        graph = stack.convergence_dependencies.graph()
        key = parser.ConvergenceNode(rsrc.id, is_update)
        if key not in graph and rsrc.replaces is not None:
            # This resource replaces old one and is not needed in
            # current traversal. You need to mark the resource as
            # DELETED so that it gets cleaned up in purge_db.
            values = {'action': rsrc.DELETE}
            db_api.resource_update_and_save(stack.context, rsrc.id, values)
            # The old resource might be in the graph (a rollback case);
            # just re-trigger it.
            check_resource.retrigger_check_resource(stack.context,
                                                    rsrc.replaces, stack)

    @context.request_context
    @log_exceptions
    def check_resource(self, cnxt, resource_id, current_traversal, data,
                       is_update, adopt_stack_data, converge=False):
        """Process a node in the dependency graph.

        The node may be associated with either an update or a cleanup of its
        associated resource.
        """
        in_data = sync_point.deserialize_input_data(data)
        resource_data = node_data.load_resources_data(in_data if is_update
                                                      else {})
        rsrc, stk_defn, stack = check_resource.load_resource(cnxt, resource_id,
                                                             resource_data,
                                                             current_traversal,
                                                             is_update)

        if rsrc is None:
            return

        rsrc.converge = converge

        msg_queue = eventlet.queue.LightQueue()
        try:
            self.thread_group_mgr.add_msg_queue(stack.id, msg_queue)
            cr = check_resource.CheckResource(self.engine_id,
                                              self._rpc_client,
                                              self.thread_group_mgr,
                                              msg_queue, in_data)
            if current_traversal != stack.current_traversal:
                LOG.debug('[%s] Traversal cancelled; re-trigerring.',
                          current_traversal)
                self._retrigger_replaced(is_update, rsrc, stack, cr)
            else:
                cr.check(cnxt, resource_id, current_traversal, resource_data,
                         is_update, adopt_stack_data, rsrc, stack)
        finally:
            self.thread_group_mgr.remove_msg_queue(None,
                                                   stack.id, msg_queue)

    @context.request_context
    @log_exceptions
    def cancel_check_resource(self, cnxt, stack_id):
        """Cancel check_resource for given stack.

        All the workers running for the given stack will be
        cancelled.
        """
        _cancel_check_resource(stack_id, self.engine_id, self.thread_group_mgr)


def _stop_traversal(stack):
    old_trvsl = stack.current_traversal
    updated = _update_current_traversal(stack)
    if not updated:
        LOG.warning("Failed to update stack %(name)s with new "
                    "traversal, aborting stack cancel", stack.name)
        return

    reason = 'Stack %(action)s cancelled' % {'action': stack.action}
    updated = stack.state_set(stack.action, stack.FAILED, reason)
    if not updated:
        LOG.warning("Failed to update stack %(name)s status "
                    "to %(action)s_%(state)s",
                    {'name': stack.name, 'action': stack.action,
                     'state': stack.FAILED})
        return

    sync_point.delete_all(stack.context, stack.id, old_trvsl)


def _update_current_traversal(stack):
    previous_traversal = stack.current_traversal
    stack.current_traversal = uuidutils.generate_uuid()
    values = {'current_traversal': stack.current_traversal}
    return stack_objects.Stack.select_and_update(
        stack.context, stack.id, values,
        exp_trvsl=previous_traversal)


def _cancel_check_resource(stack_id, engine_id, tgm):
    LOG.debug('Cancelling workers for stack [%s] in engine [%s]',
              stack_id, engine_id)
    tgm.send(stack_id, rpc_api.THREAD_CANCEL)


def _wait_for_cancellation(stack, wait=5):
    # give enough time to wait till cancel is completed
    retries = CANCEL_RETRIES
    while retries > 0:
        retries -= 1
        eventlet.sleep(wait)
        engines = db_api.engine_get_all_locked_by_stack(
            stack.context, stack.id)
        if not engines:
            return True

    return False


def _cancel_workers(stack, tgm, local_engine_id, rpc_client):
    engines = db_api.engine_get_all_locked_by_stack(stack.context, stack.id)

    if not engines:
        return True

    # cancel workers running locally
    if local_engine_id in engines:
        _cancel_check_resource(stack.id, local_engine_id, tgm)
        engines.remove(local_engine_id)

    # cancel workers on remote engines
    for engine_id in engines:
        rpc_client.cancel_check_resource(stack.context, stack.id, engine_id)

    return _wait_for_cancellation(stack)



# Copyright 2018 Ericsson
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutronclient.common import exceptions
from neutronclient.v2_0 import client as neutronclient

from heat.common import template_format
from heat.engine.clients.os import neutron
from heat.engine import scheduler
from heat.tests import common
from heat.tests import utils

try:
    from networking_l2gw.services.l2gateway.exceptions import (
        L2GatewaySegmentationRequired)  # noqa
except ImportError:
    class L2GatewaySegmentationRequired(exceptions.NeutronException):
        message = ("L2 gateway segmentation id must be consistent for all "
                   "the interfaces")


class NeutronL2GatewayTest(common.HeatTestCase):
    test_template = '''
    heat_template_version: queens
    description: Template to test L2Gateway Neutron resource
    resources:
      l2gw:
        type: OS::Neutron::L2Gateway
        properties:
          name: L2GW01
          devices:
            - device_name: switch01
              interfaces:
                - name: eth0
                - name: eth1
    '''

    test_template_update = '''
    heat_template_version: queens
    description: Template to test L2Gateway Neutron resource
    resources:
      l2gw:
        type: OS::Neutron::L2Gateway
        properties:
          name: L2GW01
          devices:
            - device_name: switch01
              interfaces:
                - name: eth0
                - name: eth1
            - device_name: switch02
              interfaces:
                - name: eth5
                - name: eth6
    '''

    test_template_with_seg = '''
    heat_template_version: queens
    description: Template to test L2Gateway Neutron resource
    resources:
      l2gw:
        type: OS::Neutron::L2Gateway
        properties:
          name: L2GW01
          devices:
            - device_name: switch01
              interfaces:
                - name: eth0
                  segmentation_id:
                    - 101
                    - 102
                    - 103
                - name: eth1
                  segmentation_id:
                    - 101
                    - 102
                    - 103
    '''

    test_template_invalid_seg = '''
    heat_template_version: queens
    description: Template to test L2Gateway Neutron resource
    resources:
      l2gw:
        type: OS::Neutron::L2Gateway
        properties:
          name: L2GW01
          devices:
            - device_name: switch01
              interfaces:
                - name: eth0
                  segmentation_id:
                    - 101
                    - 102
                    - 103
                - name: eth1
    '''

    mock_create_req = {
        "l2_gateway": {
            "name": "L2GW01",
            "devices": [{
                "device_name": "switch01",
                "interfaces": [
                    {"name": "eth0"},
                    {"name": "eth1"}
                ]
            }]
        }}
    mock_create_reply = {
        "l2_gateway": {
            "name": "L2GW01",
            "id": "d3590f37-b072-4358-9719-71964d84a31c",
            "tenant_id": "7ea656c7c9b8447494f33b0bc741d9e6",
            "devices": [{
                "device_name": "switch01",
                "interfaces": [
                    {"name": "eth0"},
                    {"name": "eth1"}
                ]
            }]
        }}

    mock_update_req = {
        "l2_gateway": {
            "devices": [{
                "device_name": "switch01",
                "interfaces": [
                    {"name": "eth0"},
                    {"name": "eth1"}
                    ]
            }, {
                "device_name": "switch02",
                "interfaces": [
                    {"name": "eth5"},
                    {"name": "eth6"}]
                }]
        }}

    mock_update_reply = {
        "l2_gateway": {
            "name": "L2GW01",
            "id": "d3590f37-b072-4358-9719-71964d84a31c",
            "tenant_id": "7ea656c7c9b8447494f33b0bc741d9e6",
            "devices": [{
                "device_name": "switch01",
                "interfaces": [
                    {"name": "eth0"},
                    {"name": "eth1"}]
            }, {
                "device_name": "switch02",
                "interfaces": [
                    {"name": "eth5"},
                    {"name": "eth6"}]
                }]
        }}

    mock_create_with_seg_req = {
        "l2_gateway": {
            "name": "L2GW01",
            "devices": [{
                "device_name": "switch01",
                "interfaces": [
                    {"name": "eth0",
                     "segmentation_id": [101, 102, 103]},
                    {"name": "eth1",
                     "segmentation_id": [101, 102, 103]}
                ]
            }]
        }}
    mock_create_with_seg_reply = {
        "l2_gateway": {
            "name": "L2GW01",
            "id": "d3590f37-b072-4358-9719-71964d84a31c",
            "tenant_id": "7ea656c7c9b8447494f33b0bc741d9e6",
            "devices": [{
                "device_name": "switch01",
                "interfaces": [
                    {"name": "eth0",
                     "segmentation_id": ["101", "102", "103"]},
                    {"name": "eth1",
                     "segmentation_id": ["101", "102", "103"]}
                ]
            }]
        }}

    mock_create_invalid_seg_req = {
        "l2_gateway": {
            "name": "L2GW01",
            "devices": [{
                "device_name": "switch01",
                "interfaces": [
                    {"name": "eth0",
                     "segmentation_id": [101, 102, 103]},
                    {"name": "eth1"}
                ]
            }]
        }}
    mock_create_invalid_seg_reply = {
        "l2_gateway": {
            "name": "L2GW01",
            "id": "d3590f37-b072-4358-9719-71964d84a31c",
            "tenant_id": "7ea656c7c9b8447494f33b0bc741d9e6",
            "devices": [{
                "device_name": "switch01",
                "interfaces": [
                    {"name": "eth0",
                     "segmentation_id": ["101", "102", "103"]},
                    {"name": "eth1"}
                ]
            }]
        }}

    def setUp(self):
        super(NeutronL2GatewayTest, self).setUp()
        self.mockclient = mock.MagicMock()
        self.patchobject(neutronclient, 'Client', return_value=self.mockclient)

        self.patchobject(neutron.NeutronClientPlugin, 'has_extension',
                         return_value=True)

    def _create_l2_gateway(self, hot, reply):
        # stack create
        self.mockclient.create_l2_gateway.return_value = reply
        self.mockclient.show_l2_gateway.return_value = reply
        template = template_format.parse(hot)
        self.stack = utils.parse_stack(template)
        scheduler.TaskRunner(self.stack.create)()
        self.l2gw_resource = self.stack['l2gw']

    def test_l2_gateway_create(self):
        self._create_l2_gateway(self.test_template, self.mock_create_reply)
        self.assertIsNone(self.l2gw_resource.validate())
        self.assertEqual((self.l2gw_resource.CREATE,
                         self.l2gw_resource.COMPLETE),
                         self.l2gw_resource.state)
        self.assertEqual('d3590f37-b072-4358-9719-71964d84a31c',
                         self.l2gw_resource.FnGetRefId())
        self.mockclient.create_l2_gateway.assert_called_once_with(
            self.mock_create_req)

    def test_l2_gateway_update(self):
        self._create_l2_gateway(self.test_template, self.mock_create_reply)
        # update l2_gateway with 2nd device
        self.mockclient.update_l2_gateway.return_value = self.mock_update_reply
        self.mockclient.show_l2_gateway.return_value = self.mock_update_reply
        updated_tmpl = template_format.parse(self.test_template_update)
        updated_stack = utils.parse_stack(updated_tmpl)
        self.stack.update(updated_stack)
        ud_l2gw_resource = self.stack['l2gw']
        self.assertIsNone(ud_l2gw_resource.validate())
        self.assertEqual((ud_l2gw_resource.UPDATE, ud_l2gw_resource.COMPLETE),
                         ud_l2gw_resource.state)
        self.assertEqual('d3590f37-b072-4358-9719-71964d84a31c',
                         ud_l2gw_resource.FnGetRefId())
        self.mockclient.update_l2_gateway.assert_called_once_with(
            'd3590f37-b072-4358-9719-71964d84a31c',
            self.mock_update_req)

    def test_l2_gateway_create_with_seg(self):
        # test with segmentation_id in template
        self._create_l2_gateway(self.test_template_with_seg,
                                self.mock_create_with_seg_reply)
        self.assertIsNone(self.l2gw_resource.validate())
        self.assertEqual((self.l2gw_resource.CREATE,
                         self.l2gw_resource.COMPLETE),
                         self.l2gw_resource.state)
        self.assertEqual('d3590f37-b072-4358-9719-71964d84a31c',
                         self.l2gw_resource.FnGetRefId())
        self.mockclient.create_l2_gateway.assert_called_once_with(
            self.mock_create_with_seg_req)

    def test_l2_gateway_create_invalid_seg(self):
        # test failure when segmentation_id is not consistent across
        # all interfaces
        self.mockclient.create_l2_gateway.side_effect = (
            L2GatewaySegmentationRequired())
        template = template_format.parse(self.test_template_invalid_seg)
        self.stack = utils.parse_stack(template)
        scheduler.TaskRunner(self.stack.create)()
        self.l2gw_resource = self.stack['l2gw']
        self.assertIsNone(self.l2gw_resource.validate())
        self.assertEqual(
            'Resource CREATE failed: '
            'L2GatewaySegmentationRequired: resources.l2gw: '
            'L2 gateway segmentation id must be consistent for '
            'all the interfaces',
            self.stack.status_reason)
        self.assertEqual((self.l2gw_resource.CREATE,
                         self.l2gw_resource.FAILED),
                         self.l2gw_resource.state)
        self.mockclient.create_l2_gateway.assert_called_once_with(
            self.mock_create_invalid_seg_req)

    def test_l2_gateway_delete(self):
        self._create_l2_gateway(self.test_template, self.mock_create_reply)
        self.stack.delete()
        self.mockclient.delete_l2_gateway.assert_called_with(
            'd3590f37-b072-4358-9719-71964d84a31c')



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common.i18n import _
from heat.engine import constraints
from heat.engine import properties
from heat.engine import resource
from heat.engine import support


class SecurityService(resource.Resource):
    """A resource that implements security service of Manila.

    A security_service is a set of options that defines a security domain
    for a particular shared filesystem protocol, such as an
    Active Directory domain or a Kerberos domain.
    """

    support_status = support.SupportStatus(version='5.0.0')

    PROPERTIES = (
        NAME, TYPE, DNS_IP, SERVER, DOMAIN, USER,
        PASSWORD, DESCRIPTION
    ) = (
        'name', 'type', 'dns_ip', 'server', 'domain', 'user',
        'password', 'description'
    )

    properties_schema = {
        NAME: properties.Schema(
            properties.Schema.STRING,
            _('Security service name.'),
            update_allowed=True
        ),
        TYPE: properties.Schema(
            properties.Schema.STRING,
            _('Security service type.'),
            required=True,
            constraints=[
                constraints.AllowedValues(['ldap', 'kerberos',
                                           'active_directory'])
            ]
        ),
        DNS_IP: properties.Schema(
            properties.Schema.STRING,
            _('DNS IP address used inside tenant\'s network.'),
            update_allowed=True
        ),
        SERVER: properties.Schema(
            properties.Schema.STRING,
            _('Security service IP address or hostname.'),
            update_allowed=True
        ),
        DOMAIN: properties.Schema(
            properties.Schema.STRING,
            _('Security service domain.'),
            update_allowed=True
        ),
        USER: properties.Schema(
            properties.Schema.STRING,
            _('Security service user or group used by tenant.'),
            update_allowed=True
        ),
        PASSWORD: properties.Schema(
            properties.Schema.STRING,
            _('Password used by user.'),
            update_allowed=True
        ),
        DESCRIPTION: properties.Schema(
            properties.Schema.STRING,
            _('Security service description.'),
            update_allowed=True
        )
    }

    default_client_name = 'manila'

    entity = 'security_services'

    def handle_create(self):
        args = dict((k, v) for k, v in self.properties.items()
                    if v is not None)
        security_service = self.client().security_services.create(**args)
        self.resource_id_set(security_service.id)

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        if prop_diff:
            self.client().security_services.update(self.resource_id,
                                                   **prop_diff)


def resource_mapping():
    return {
        'OS::Manila::SecurityService': SecurityService
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common import exception
from heat.common.i18n import _
from heat.engine import attributes
from heat.engine import constraints
from heat.engine import properties
from heat.engine.resources.openstack.octavia import octavia_base
from heat.engine import support
from heat.engine import translation


class Pool(octavia_base.OctaviaBase):
    """A resource for managing Octavia Pools.

    This resources manages octavia LBaaS Pools, which represent a group
    of nodes. Pools define the subnet where nodes reside, balancing algorithm,
    and the nodes themselves.
    """

    PROPERTIES = (
        ADMIN_STATE_UP, DESCRIPTION, SESSION_PERSISTENCE, NAME,
        LB_ALGORITHM, LISTENER, LOADBALANCER, PROTOCOL,
        SESSION_PERSISTENCE_TYPE, SESSION_PERSISTENCE_COOKIE_NAME,
        TLS_ENABLED,
    ) = (
        'admin_state_up', 'description', 'session_persistence', 'name',
        'lb_algorithm', 'listener', 'loadbalancer', 'protocol',
        'type', 'cookie_name', 'tls_enabled',
    )

    SESSION_PERSISTENCE_TYPES = (
        SOURCE_IP, HTTP_COOKIE, APP_COOKIE
    ) = (
        'SOURCE_IP', 'HTTP_COOKIE', 'APP_COOKIE'
    )

    SUPPORTED_PROTOCOLS = (TCP, HTTP, HTTPS, TERMINATED_HTTPS, PROXY, UDP) = (
        'TCP', 'HTTP', 'HTTPS', 'TERMINATED_HTTPS', 'PROXY', 'UDP')

    ATTRIBUTES = (
        HEALTHMONITOR_ID_ATTR, LISTENERS_ATTR, MEMBERS_ATTR
    ) = (
        'healthmonitor_id', 'listeners', 'members'
    )

    properties_schema = {
        ADMIN_STATE_UP: properties.Schema(
            properties.Schema.BOOLEAN,
            _('The administrative state of this pool.'),
            default=True,
            update_allowed=True
        ),
        DESCRIPTION: properties.Schema(
            properties.Schema.STRING,
            _('Description of this pool.'),
            update_allowed=True,
            default=''
        ),
        SESSION_PERSISTENCE: properties.Schema(
            properties.Schema.MAP,
            _('Configuration of session persistence.'),
            schema={
                SESSION_PERSISTENCE_TYPE: properties.Schema(
                    properties.Schema.STRING,
                    _('Method of implementation of session '
                      'persistence feature.'),
                    required=True,
                    constraints=[constraints.AllowedValues(
                        SESSION_PERSISTENCE_TYPES
                    )]
                ),
                SESSION_PERSISTENCE_COOKIE_NAME: properties.Schema(
                    properties.Schema.STRING,
                    _('Name of the cookie, '
                      'required if type is APP_COOKIE.')
                )
            },
            update_allowed=True,
        ),
        NAME: properties.Schema(
            properties.Schema.STRING,
            _('Name of this pool.'),
            update_allowed=True
        ),
        LB_ALGORITHM: properties.Schema(
            properties.Schema.STRING,
            _('The algorithm used to distribute load between the members of '
              'the pool.'),
            required=True,
            constraints=[
                constraints.AllowedValues(['ROUND_ROBIN', 'LEAST_CONNECTIONS',
                                           'SOURCE_IP', 'SOURCE_IP_PORT']),
            ],
            update_allowed=True,
        ),
        LISTENER: properties.Schema(
            properties.Schema.STRING,
            _('Listener name or ID to be associated with this pool.'),
            constraints=[
                constraints.CustomConstraint('octavia.listener')
            ]
        ),
        LOADBALANCER: properties.Schema(
            properties.Schema.STRING,
            _('Loadbalancer name or ID to be associated with this pool.'),
            constraints=[
                constraints.CustomConstraint('octavia.loadbalancer')
            ],
        ),
        PROTOCOL: properties.Schema(
            properties.Schema.STRING,
            _('Protocol of the pool.'),
            required=True,
            constraints=[
                constraints.AllowedValues(SUPPORTED_PROTOCOLS),
            ]
        ),
        TLS_ENABLED: properties.Schema(
            properties.Schema.BOOLEAN,
            _('Enable backend member re-encryption.'),
            default=False,
            update_allowed=True,
            support_status=support.SupportStatus(version='14.0.0'),
        ),
    }

    attributes_schema = {
        HEALTHMONITOR_ID_ATTR: attributes.Schema(
            _('ID of the health monitor associated with this pool.'),
            type=attributes.Schema.STRING
        ),
        LISTENERS_ATTR: attributes.Schema(
            _('Listener associated with this pool.'),
            type=attributes.Schema.STRING
        ),
        MEMBERS_ATTR: attributes.Schema(
            _('Members associated with this pool.'),
            cache_mode=attributes.Schema.CACHE_NONE,
            type=attributes.Schema.LIST
        ),
    }

    def translation_rules(self, props):
        return [
            translation.TranslationRule(
                props,
                translation.TranslationRule.RESOLVE,
                [self.LISTENER],
                client_plugin=self.client_plugin(),
                finder='get_listener',
            ),
            translation.TranslationRule(
                props,
                translation.TranslationRule.RESOLVE,
                [self.LOADBALANCER],
                client_plugin=self.client_plugin(),
                finder='get_loadbalancer',
            ),
        ]

    def _prepare_args(self, properties):
        props = dict((k, v) for k, v in properties.items() if v is not None)
        if self.NAME not in props:
            props[self.NAME] = self.physical_resource_name()
        if self.LISTENER in props:
            props['listener_id'] = props.pop(self.LISTENER)
        if self.LOADBALANCER in props:
            props['loadbalancer_id'] = props.pop(self.LOADBALANCER)
        self._prepare_session_persistence(props)
        return props

    def _prepare_session_persistence(self, props):
        session_p = props.get(self.SESSION_PERSISTENCE)
        if session_p is not None:
            session_props = dict(
                (k, v) for k, v in session_p.items() if v is not None)
            props[self.SESSION_PERSISTENCE] = session_props

    def validate(self):
        super(Pool, self).validate()
        if (self.properties[self.LISTENER] is None and
                self.properties[self.LOADBALANCER] is None):
            raise exception.PropertyUnspecifiedError(self.LISTENER,
                                                     self.LOADBALANCER)

        if self.properties[self.SESSION_PERSISTENCE] is not None:
            session_p = self.properties[self.SESSION_PERSISTENCE]
            persistence_type = session_p[self.SESSION_PERSISTENCE_TYPE]
            if persistence_type == self.APP_COOKIE:
                if not session_p.get(self.SESSION_PERSISTENCE_COOKIE_NAME):
                    msg = (_('Property %(cookie)s is required when %(sp)s '
                             'type is set to %(app)s.') %
                           {'cookie': self.SESSION_PERSISTENCE_COOKIE_NAME,
                            'sp': self.SESSION_PERSISTENCE,
                            'app': self.APP_COOKIE})
                    raise exception.StackValidationFailed(message=msg)
            elif persistence_type == self.SOURCE_IP:
                if session_p.get(self.SESSION_PERSISTENCE_COOKIE_NAME):
                    msg = (_('Property %(cookie)s must NOT be specified when '
                             '%(sp)s type is set to %(ip)s.') %
                           {'cookie': self.SESSION_PERSISTENCE_COOKIE_NAME,
                            'sp': self.SESSION_PERSISTENCE,
                            'ip': self.SOURCE_IP})
                    raise exception.StackValidationFailed(message=msg)

    def _resource_create(self, properties):
        return self.client().pool_create(json={'pool': properties})['pool']

    def _resource_update(self, prop_diff):
        props = dict((k, v) for k, v in prop_diff.items() if v is not None)
        self._prepare_session_persistence(props)
        self.client().pool_set(self.resource_id, json={'pool': props})

    def _resource_delete(self):
        self.client().pool_delete(self.resource_id)

    def _show_resource(self):
        return self.client().pool_show(self.resource_id)


def resource_mapping():
    return {
        'OS::Octavia::Pool': Pool,
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common import exception
from heat.common.i18n import _
from heat.engine import attributes
from heat.engine import properties
from heat.engine import resource
from heat.engine import support


class Host(resource.Resource):
    """A resource to manage Blazar hosts.

    Host resource manages the physical hosts for the lease/reservation
    within OpenStack.

    # TODO(asmita): Based on an agreement with Blazar team, this resource
    class does not support updating host resource as currently Blazar does
    not support to delete existing extra_capability keys while updating host.
    Also, in near future, when Blazar team will come up with a new alternative
    API to resolve this issue, we will need to modify this class.
    """

    support_status = support.SupportStatus(version='12.0.0')

    PROPERTIES = (
        NAME, EXTRA_CAPABILITY,
    ) = (
        'name', 'extra_capability',
    )

    ATTRIBUTES = (
        HYPERVISOR_HOSTNAME, HYPERVISOR_TYPE, HYPERVISOR_VERSION,
        VCPUS, CPU_INFO, MEMORY_MB, LOCAL_GB,
        SERVICE_NAME, RESERVABLE, STATUS, TRUST_ID,
        EXTRA_CAPABILITY_ATTR, CREATED_AT, UPDATED_AT,
    ) = (
        'hypervisor_hostname', 'hypervisor_type', 'hypervisor_version',
        'vcpus', 'cpu_info', 'memory_mb', 'local_gb',
        'service_name', 'reservable', 'status', 'trust_id',
        'extra_capability', 'created_at', 'updated_at',
    )

    properties_schema = {
        NAME: properties.Schema(
            properties.Schema.STRING,
            _('The name of the host.'),
            required=True,
        ),
        EXTRA_CAPABILITY: properties.Schema(
            properties.Schema.MAP,
            _('The extra capability of the host.'),
        )
    }

    attributes_schema = {
        HYPERVISOR_HOSTNAME: attributes.Schema(
            _('The hypervisor name of the host.'),
            type=attributes.Schema.STRING,
        ),
        HYPERVISOR_TYPE: attributes.Schema(
            _('The hypervisor type the host.'),
            type=attributes.Schema.STRING,
        ),
        HYPERVISOR_VERSION: attributes.Schema(
            _('The hypervisor version of the host.'),
            type=attributes.Schema.INTEGER,
        ),
        VCPUS: attributes.Schema(
            _('The number of the VCPUs of the host.'),
            type=attributes.Schema.INTEGER,
        ),
        CPU_INFO: attributes.Schema(
            _('Information of the CPU of the host.'),
            type=attributes.Schema.MAP,
        ),
        MEMORY_MB: attributes.Schema(
            _('Megabytes of the memory of the host.'),
            type=attributes.Schema.INTEGER,
        ),
        LOCAL_GB: attributes.Schema(
            _('Gigabytes of the disk of the host.'),
            type=attributes.Schema.INTEGER,
        ),
        SERVICE_NAME: attributes.Schema(
            _('The compute service name of the host.'),
            type=attributes.Schema.STRING,
        ),
        RESERVABLE: attributes.Schema(
            _('The flag which represents whether the host is reservable '
              'or not.'),
            type=attributes.Schema.BOOLEAN,
        ),
        STATUS: attributes.Schema(
            _('The status of the host.'),
            type=attributes.Schema.STRING,
        ),
        TRUST_ID: attributes.Schema(
            _('The UUID of the trust of the host operator.'),
            type=attributes.Schema.STRING,
        ),
        EXTRA_CAPABILITY_ATTR: attributes.Schema(
            _('The extra capability of the host.'),
            type=attributes.Schema.MAP,
        ),
        CREATED_AT: attributes.Schema(
            _('The date and time when the host was created. '
              'The date and time format must be "CCYY-MM-DD hh:mm".'),
            type=attributes.Schema.STRING,
        ),
        UPDATED_AT: attributes.Schema(
            _('The date and time when the host was updated. '
              'The date and time format must be "CCYY-MM-DD hh:mm".'),
            type=attributes.Schema.STRING
        ),
    }

    default_client_name = 'blazar'

    entity = 'host'

    def _parse_extra_capability(self, args):
        if self.NAME in args[self.EXTRA_CAPABILITY]:
            # Remove "name" key if present in the extra_capability property.
            del args[self.EXTRA_CAPABILITY][self.NAME]
        args.update(args[self.EXTRA_CAPABILITY])
        args.pop(self.EXTRA_CAPABILITY)
        return args

    def handle_create(self):
        args = dict((k, v) for k, v in self.properties.items()
                    if v is not None)

        if self.EXTRA_CAPABILITY in args:
            args = self._parse_extra_capability(args)

        host = self.client_plugin().create_host(**args)
        self.resource_id_set(host['id'])

        return host['id']

    def _resolve_attribute(self, name):
        if self.resource_id is None:
            return
        host = self.client_plugin().get_host(self.resource_id)
        try:
            return host[name]
        except KeyError:
            raise exception.InvalidTemplateAttribute(resource=self.name,
                                                     key=name)


def resource_mapping():
    return {
        'OS::Blazar::Host': Host
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from heat.common import exception
from heat.common import template_format
from heat.engine.clients.os import mistral as client
from heat.engine import resource
from heat.engine.resources.openstack.mistral import external_resource
from heat.engine import scheduler
from heat.engine import template
from heat.tests import common
from heat.tests import utils

external_resource_template = """
heat_template_version: ocata
resources:
  custom:
    type: OS::Mistral::ExternalResource
    properties:
      actions:
        CREATE:
          workflow: some_workflow
          params:
            target: create_my_custom_thing
        UPDATE:
          workflow: another_workflow
        DELETE:
          workflow: yet_another_workflow
      input:
        foo1: 123
        foo2: 456
      replace_on_change_inputs:
        - foo2
"""


class FakeExecution(object):
    def __init__(self, id='1234', output='{}', state='IDLE'):
        self.id = id
        self.output = output
        self.state = state


class TestMistralExternalResource(common.HeatTestCase):
    def setUp(self):
        super(TestMistralExternalResource, self).setUp()
        self.ctx = utils.dummy_context()
        tmpl = template_format.parse(external_resource_template)
        self.stack = utils.parse_stack(tmpl, stack_name='test_stack')

        resource_defns = self.stack.t.resource_definitions(self.stack)
        self.rsrc_defn = resource_defns['custom']

        self.mistral = mock.Mock()
        self.patchobject(external_resource.MistralExternalResource,
                         'client',
                         return_value=self.mistral)
        self.patchobject(client.MistralClientPlugin, '_create')
        self.client = client.MistralClientPlugin(self.ctx)

    def _create_resource(self, name, snippet, stack,
                         output='{}',
                         get_state='SUCCESS'):
        execution = external_resource.MistralExternalResource(name,
                                                              snippet,
                                                              stack)
        self.mistral.executions.get.return_value = (
            FakeExecution('test_stack-execution-b5fiekfci3yc',
                          output,
                          get_state))
        self.mistral.executions.create.return_value = (
            FakeExecution('test_stack-execution-b5fiekfci3yc'))
        return execution

    def test_create(self):
        execution = self._create_resource('execution',
                                          self.rsrc_defn,
                                          self.stack)
        scheduler.TaskRunner(execution.create)()
        expected_state = (execution.CREATE, execution.COMPLETE)
        self.assertEqual(expected_state, execution.state)
        self.assertEqual('test_stack-execution-b5fiekfci3yc',
                         execution.resource_id)

    def test_create_with_resource_id_output(self):
        output = '{"resource_id": "my-fake-resource-id"}'
        execution = self._create_resource('execution',
                                          self.rsrc_defn,
                                          self.stack,
                                          output)
        scheduler.TaskRunner(execution.create)()
        expected_state = (execution.CREATE, execution.COMPLETE)
        self.assertEqual(expected_state, execution.state)
        self.assertEqual('my-fake-resource-id',
                         execution.resource_id)

    def test_replace_on_change(self):
        execution = self._create_resource('execution',
                                          self.rsrc_defn,
                                          self.stack)
        scheduler.TaskRunner(execution.create)()
        expected_state = (execution.CREATE, execution.COMPLETE)
        self.assertEqual(expected_state, execution.state)

        tmpl = template_format.parse(external_resource_template)
        tmpl['resources']['custom']['properties']['input']['foo2'] = '4567'
        res_defns = template.Template(tmpl).resource_definitions(self.stack)
        new_custom_defn = res_defns['custom']
        self.assertRaises(resource.UpdateReplace,
                          scheduler.TaskRunner(execution.update,
                                               new_custom_defn))

    def test_create_failed(self):
        execution = self._create_resource('execution',
                                          self.rsrc_defn,
                                          self.stack,
                                          get_state='ERROR')
        self.assertRaises(exception.ResourceFailure,
                          scheduler.TaskRunner(execution.create))
        expected_state = (execution.CREATE, execution.FAILED)
        self.assertEqual(expected_state, execution.state)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json

from heat.common import environment_util as env_util
from heat.common import exception
from heat.engine import parameters
from heat.tests import common
from heat.tests import utils


class TestEnvironmentUtil(common.HeatTestCase):

    def test_empty_merge_strategies(self):
        merge_strategies = {}
        param_strategy = env_util.get_param_merge_strategy(merge_strategies,
                                                           'param1')
        self.assertEqual(env_util.OVERWRITE, param_strategy)

    def test_default_merge_strategy(self):
        merge_strategies = {'default': 'deep_merge'}
        param_strategy = env_util.get_param_merge_strategy(merge_strategies,
                                                           'param1')
        self.assertEqual(env_util.DEEP_MERGE, param_strategy)

    def test_param_sepcific_merge_strategy(self):
        merge_strategies = {'default': 'merge',
                            'param1': 'deep_merge'}
        param_strategy = env_util.get_param_merge_strategy(merge_strategies,
                                                           'param1')
        self.assertEqual(env_util.DEEP_MERGE, param_strategy)

    def test_wrong_param_strategy(self):
        merge_strategies = {'default': 'merge',
                            'param1': 'unknown'}
        param_strategy = env_util.get_param_merge_strategy(merge_strategies,
                                                           'param1')
        self.assertEqual(env_util.MERGE, param_strategy)

    def test_merge_startegies_none(self):
        merge_strategies = None
        param_strategy = env_util.get_param_merge_strategy(merge_strategies,
                                                           'param1')
        self.assertEqual(env_util.OVERWRITE, param_strategy)


class TestMergeEnvironments(common.HeatTestCase):
    def setUp(self):
        super(TestMergeEnvironments, self).setUp()
        self.ctx = utils.dummy_context(tenant_id='stack_service_test_tenant')
        # Setup
        self.params = {'parameters': {},
                       'resource_registry': {},
                       'parameter_defaults': {}}

        self.env_1 = {'parameters': {
            'str_value1': "string1",
            'str_value2': "string2",
            'del_lst_value1': '1,2',
            'del_lst_value2': '3,4',
            'lst_value1': [1, 2],
            'json_value1': {"1": ["str1", "str2"]},
            'json_value2': {"2": ["test1", "test2"]}},
            'resource_registry': {
                'test::R1': "OS::Heat::RandomString",
                'test::R2': "BROKEN"},
            'parameter_defaults': {
                'lst_value2': [3, 4]}}
        self.env_2 = {'parameters': {
            'str_value1': "string3",
            'str_value2': "string4",
            'del_lst_value1': '5,6',
            'del_lst_value2': '7,8',
            'lst_value1': [5, 6],
            'json_value1': {"3": ["str3", "str4"]},
            'json_value2': {"4": ["test3", "test4"]}},
            'resource_registry': {
                'test::R2': "OS::Heat::None"},
            'parameter_defaults': {
                'lst_value2': [7, 8]}}

        self.env_3 = {'parameters': {
            'lst_value1': [9, 10],
            'json_value1': {"5": ["str5"]}}}

        self.env_4 = {'parameter_defaults': {
            'lst_value2': [9, 10]}}

        self.param_schemata = {
            'str_value1': parameters.Schema(parameters.Schema.STRING),
            'str_value2': parameters.Schema(parameters.Schema.STRING),
            'del_lst_value1': parameters.Schema(parameters.Schema.LIST),
            'del_lst_value2': parameters.Schema(parameters.Schema.LIST),
            'lst_value1': parameters.Schema(parameters.Schema.LIST,
                                            default=[7]),
            'lst_value2': parameters.Schema(parameters.Schema.LIST),
            'json_value1': parameters.Schema(parameters.Schema.MAP),
            'json_value2': parameters.Schema(parameters.Schema.MAP)}

    def test_merge_envs_with_param_default_merge_strategy(self):
        files = {'env_1': json.dumps(self.env_1),
                 'env_2': json.dumps(self.env_2)}
        environment_files = ['env_1', 'env_2']

        # Test
        env_util.merge_environments(environment_files, files, self.params,
                                    self.param_schemata)

        # Verify
        expected = {'parameters': {
                    'json_value1': {'3': ['str3', 'str4']},
                    'json_value2': {'4': ['test3', 'test4']},
                    'del_lst_value1': '5,6',
                    'del_lst_value2': '7,8',
                    'lst_value1': [5, 6],
                    'str_value1': 'string3',
                    'str_value2': 'string4'},
                    'resource_registry': {
                        'test::R1': "OS::Heat::RandomString",
                        'test::R2': "OS::Heat::None"},
                    'parameter_defaults': {
                        'lst_value2': [7, 8]}}
        self.assertEqual(expected, self.params)

    def test_merge_envs_with_specified_default(self):
        merge_strategies = {'default': 'deep_merge'}
        self.env_1['parameter_merge_strategies'] = merge_strategies
        files = {'env_1': json.dumps(self.env_1),
                 'env_2': json.dumps(self.env_2)}
        environment_files = ['env_1', 'env_2']

        # Test
        env_util.merge_environments(environment_files, files, self.params,
                                    self.param_schemata)

        # Verify
        expected = {'parameters': {
                    'json_value1': {'3': ['str3', 'str4'],
                                    '1': ['str1', 'str2']},  # added
                    'json_value2': {'4': ['test3', 'test4'],
                                    '2': ['test1', 'test2']},
                    'del_lst_value1': '1,2,5,6',
                    'del_lst_value2': '3,4,7,8',
                    'lst_value1': [1, 2, 5, 6],  # added
                    'str_value1': 'string1string3',
                    'str_value2': 'string2string4'},
                    'resource_registry': {
                        'test::R1': "OS::Heat::RandomString",
                        'test::R2': "OS::Heat::None"},
                    'parameter_defaults': {
                        'lst_value2': [3, 4, 7, 8]}}
        self.assertEqual(expected, self.params)

    def test_merge_envs_with_param_specific_merge_strategy(self):
        merge_strategies = {
            'default': 'overwrite',
            'lst_value1': 'merge',
            'lst_value2': 'merge',
            'json_value1': 'deep_merge'}

        self.env_1['parameter_merge_strategies'] = merge_strategies

        files = {'env_1': json.dumps(self.env_1),
                 'env_2': json.dumps(self.env_2)}
        environment_files = ['env_1', 'env_2']

        # Test
        env_util.merge_environments(environment_files, files, self.params,
                                    self.param_schemata)

        # Verify
        expected = {'parameters': {
                    'json_value1': {'3': ['str3', 'str4'],
                                    '1': ['str1', 'str2']},  # added
                    'json_value2': {'4': ['test3', 'test4']},
                    'del_lst_value1': '5,6',
                    'del_lst_value2': '7,8',
                    'lst_value1': [1, 2, 5, 6],  # added
                    'str_value1': 'string3',
                    'str_value2': 'string4'},
                    'resource_registry': {
                        'test::R1': 'OS::Heat::RandomString',
                        'test::R2': 'OS::Heat::None'},
                    'parameter_defaults': {
                        'lst_value2': [3, 4, 7, 8]}}
        self.assertEqual(expected, self.params)

    def test_merge_envs_with_param_conflicting_merge_strategy(self):
        merge_strategies = {
            'default': "overwrite",
            'lst_value1': "merge",
            'json_value1': "deep_merge"}

        env3_merge_strategies = {
            'default': "overwrite",
            'lst_value1': "deep_merge",
            'json_value1': "merge"}

        self.env_1['parameter_merge_strategies'] = merge_strategies
        self.env_3['parameter_merge_strategies'] = env3_merge_strategies

        files = {'env_1': json.dumps(self.env_1),
                 'env_2': json.dumps(self.env_2),
                 'env_3': json.dumps(self.env_3)}

        environment_files = ['env_1', 'env_2', 'env_3']

        # Test
        self.assertRaises(exception.ConflictingMergeStrategyForParam,
                          env_util.merge_environments,
                          environment_files, files,
                          self.params, self.param_schemata)

    def test_merge_envs_with_param_defaults_conflicting_merge_strategy(self):
        merge_strategies = {
            'default': "overwrite",
            'lst_value2': "merge"}

        env4_merge_strategies = {
            'default': "overwrite",
            'lst_value2': "overwrite"}

        self.env_1['parameter_merge_strategies'] = merge_strategies
        self.env_4['parameter_merge_strategies'] = env4_merge_strategies

        files = {'env_1': json.dumps(self.env_1),
                 'env_2': json.dumps(self.env_2),
                 'env_4': json.dumps(self.env_4)}

        environment_files = ['env_1', 'env_2', 'env_4']

        # Test
        self.assertRaises(exception.ConflictingMergeStrategyForParam,
                          env_util.merge_environments,
                          environment_files, files,
                          self.params, self.param_schemata)

    def test_merge_environments_no_env_files(self):
        files = {'env_1': json.dumps(self.env_1)}

        # Test - Should ignore env_1 in files
        env_util.merge_environments(None, files, self.params,
                                    self.param_schemata)

        # Verify
        expected = {'parameters': {},
                    'resource_registry': {},
                    'parameter_defaults': {}}

        self.assertEqual(expected, self.params)

    def test_merge_envs_with_zeros(self):
        env1 = {'parameter_defaults': {'value1': 1}}
        env2 = {'parameter_defaults': {'value1': 0}}
        files = {'env_1': json.dumps(env1),
                 'env_2': json.dumps(env2)}
        environment_files = ['env_1', 'env_2']

        param_schemata = {
            'value1': parameters.Schema(parameters.Schema.NUMBER)}
        env_util.merge_environments(environment_files, files, self.params,
                                    param_schemata)

        self.assertEqual({'value1': 0}, self.params['parameter_defaults'])

    def test_merge_envs_with_zeros_in_maps(self):
        env1 = {'parameter_defaults': {'value1': {'foo': 1}},
                'parameter_merge_strategies': {'value1': 'deep_merge'}}
        env2 = {'parameter_defaults': {'value1': {'foo': 0}}}
        files = {'env_1': json.dumps(env1),
                 'env_2': json.dumps(env2)}
        environment_files = ['env_1', 'env_2']

        param_schemata = {
            'value1': parameters.Schema(parameters.Schema.MAP)}
        env_util.merge_environments(environment_files, files, self.params,
                                    param_schemata)

        self.assertEqual({'value1': {'foo': 0}},
                         self.params['parameter_defaults'])



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common import exception
from heat.common.i18n import _


class EntityMatchNotFound(exception.HeatException):
    msg_fmt = _("No %(entity)s matching %(args)s.")

    def __init__(self, entity=None, args=None, **kwargs):
        super(EntityMatchNotFound, self).__init__(entity=entity, args=args,
                                                  **kwargs)


class EntityUniqueMatchNotFound(EntityMatchNotFound):
    msg_fmt = _("No %(entity)s unique match found for %(args)s.")


class InterfaceNotFound(exception.HeatException):
    msg_fmt = _("No network interface found for server %(id)s.")



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common import exception
from heat.common.i18n import _
from heat.common import password_gen
from heat.engine import attributes
from heat.engine import constraints
from heat.engine import properties
from heat.engine import resource
from heat.engine import support
from heat.engine import translation


class RandomString(resource.Resource):
    """A resource which generates a random string.

    This is useful for configuring passwords and secrets on services. Random
    string can be generated from specified character sequences, which means
    that all characters will be randomly chosen from specified sequences, or
    with some classes, e.g. letterdigits, which means that all character will
    be randomly chosen from union of ascii letters and digits. Output string
    will be randomly generated string with specified length (or with length of
    32, if length property doesn't specified).
    """

    support_status = support.SupportStatus(version='2014.1')

    PROPERTIES = (
        LENGTH, SEQUENCE, CHARACTER_CLASSES, CHARACTER_SEQUENCES,
        SALT,
    ) = (
        'length', 'sequence', 'character_classes', 'character_sequences',
        'salt',
    )

    _CHARACTER_CLASSES_KEYS = (
        CHARACTER_CLASSES_CLASS, CHARACTER_CLASSES_MIN,
    ) = (
        'class', 'min',
    )

    _CHARACTER_SEQUENCES = (
        CHARACTER_SEQUENCES_SEQUENCE, CHARACTER_SEQUENCES_MIN,
    ) = (
        'sequence', 'min',
    )

    ATTRIBUTES = (
        VALUE,
    ) = (
        'value',
    )

    properties_schema = {
        LENGTH: properties.Schema(
            properties.Schema.INTEGER,
            _('Length of the string to generate.'),
            default=32,
            constraints=[
                constraints.Range(1, 512),
            ]
        ),
        SEQUENCE: properties.Schema(
            properties.Schema.STRING,
            _('Sequence of characters to build the random string from.'),
            constraints=[
                constraints.AllowedValues(password_gen.CHARACTER_CLASSES),
            ],
            support_status=support.SupportStatus(
                status=support.HIDDEN,
                version='5.0.0',
                previous_status=support.SupportStatus(
                    status=support.DEPRECATED,
                    message=_('Use property %s.') % CHARACTER_CLASSES,
                    version='2014.2'
                )
            )
        ),
        CHARACTER_CLASSES: properties.Schema(
            properties.Schema.LIST,
            _('A list of character class and their constraints to generate '
              'the random string from.'),
            schema=properties.Schema(
                properties.Schema.MAP,
                schema={
                    CHARACTER_CLASSES_CLASS: properties.Schema(
                        properties.Schema.STRING,
                        (_('A character class and its corresponding %(min)s '
                           'constraint to generate the random string from.')
                         % {'min': CHARACTER_CLASSES_MIN}),
                        constraints=[
                            constraints.AllowedValues(
                                password_gen.CHARACTER_CLASSES),
                        ],
                        default=password_gen.LETTERS_DIGITS),
                    CHARACTER_CLASSES_MIN: properties.Schema(
                        properties.Schema.INTEGER,
                        _('The minimum number of characters from this '
                          'character class that will be in the generated '
                          'string.'),
                        default=1,
                        constraints=[
                            constraints.Range(1, 512),
                        ]
                    )
                }
            ),
            # add defaults for backward compatibility
            default=[{CHARACTER_CLASSES_CLASS: password_gen.LETTERS_DIGITS,
                      CHARACTER_CLASSES_MIN: 1}]

        ),
        CHARACTER_SEQUENCES: properties.Schema(
            properties.Schema.LIST,
            _('A list of character sequences and their constraints to '
              'generate the random string from.'),
            schema=properties.Schema(
                properties.Schema.MAP,
                schema={
                    CHARACTER_SEQUENCES_SEQUENCE: properties.Schema(
                        properties.Schema.STRING,
                        _('A character sequence and its corresponding %(min)s '
                          'constraint to generate the random string '
                          'from.') % {'min': CHARACTER_SEQUENCES_MIN},
                        required=True),
                    CHARACTER_SEQUENCES_MIN: properties.Schema(
                        properties.Schema.INTEGER,
                        _('The minimum number of characters from this '
                          'sequence that will be in the generated '
                          'string.'),
                        default=1,
                        constraints=[
                            constraints.Range(1, 512),
                        ]
                    )
                }
            )
        ),
        SALT: properties.Schema(
            properties.Schema.STRING,
            _('Value which can be set or changed on stack update to trigger '
              'the resource for replacement with a new random string. The '
              'salt value itself is ignored by the random generator.')
        ),
    }

    attributes_schema = {
        VALUE: attributes.Schema(
            _('The random string generated by this resource. This value is '
              'also available by referencing the resource.'),
            cache_mode=attributes.Schema.CACHE_NONE,
            type=attributes.Schema.STRING
        ),
    }

    def translation_rules(self, props):
        if props.get(self.SEQUENCE):
            return [
                translation.TranslationRule(
                    props,
                    translation.TranslationRule.ADD,
                    [self.CHARACTER_CLASSES],
                    [{self.CHARACTER_CLASSES_CLASS: props.get(
                        self.SEQUENCE),
                        self.CHARACTER_CLASSES_MIN: 1}]),
                translation.TranslationRule(
                    props,
                    translation.TranslationRule.DELETE,
                    [self.SEQUENCE]
                )
            ]

    def _generate_random_string(self, char_sequences, char_classes, length):
        seq_mins = [
            password_gen.special_char_class(
                char_seq[self.CHARACTER_SEQUENCES_SEQUENCE],
                char_seq[self.CHARACTER_SEQUENCES_MIN])
            for char_seq in char_sequences]
        char_class_mins = [
            password_gen.named_char_class(
                char_class[self.CHARACTER_CLASSES_CLASS],
                char_class[self.CHARACTER_CLASSES_MIN])
            for char_class in char_classes]

        return password_gen.generate_password(length,
                                              seq_mins + char_class_mins)

    def validate(self):
        super(RandomString, self).validate()
        char_sequences = self.properties[self.CHARACTER_SEQUENCES]
        char_classes = self.properties[self.CHARACTER_CLASSES]

        def char_min(char_dicts, min_prop):
            if char_dicts:
                return sum(char_dict[min_prop] for char_dict in char_dicts)
            return 0

        length = self.properties[self.LENGTH]
        min_length = (char_min(char_sequences, self.CHARACTER_SEQUENCES_MIN) +
                      char_min(char_classes, self.CHARACTER_CLASSES_MIN))
        if min_length > length:
            msg = _("Length property cannot be smaller than combined "
                    "character class and character sequence minimums")
            raise exception.StackValidationFailed(message=msg)

    def handle_create(self):
        char_sequences = self.properties[self.CHARACTER_SEQUENCES] or []
        char_classes = self.properties[self.CHARACTER_CLASSES] or []
        length = self.properties[self.LENGTH]

        random_string = self._generate_random_string(char_sequences,
                                                     char_classes,
                                                     length)
        self.data_set('value', random_string, redact=True)
        self.resource_id_set(self.physical_resource_name())

    def _resolve_attribute(self, name):
        if name == self.VALUE:
            return self.data().get(self.VALUE)

    def get_reference_id(self):
        if self.resource_id is not None:
            return self.data().get('value')
        else:
            return str(self.name)


def resource_mapping():
    return {
        'OS::Heat::RandomString': RandomString,
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common.i18n import _
from heat.engine import constraints
from heat.engine import properties
from heat.engine.resources.openstack.neutron import neutron
from heat.engine import support
from heat.engine import translation


class FlowClassifier(neutron.NeutronResource):
    """"Heat Template Resource for networking-sfc flow-classifier.

    This resource used to select the traffic that can access the service chain.
    Traffic that matches any flow classifier will be directed to the first
    port in the chain.
    """

    support_status = support.SupportStatus(
        version='8.0.0',
        status=support.UNSUPPORTED)

    PROPERTIES = (
        NAME, DESCRIPTION, PROTOCOL, ETHERTYPE,
        SOURCE_IP_PREFIX, DESTINATION_IP_PREFIX, SOURCE_PORT_RANGE_MIN,
        SOURCE_PORT_RANGE_MAX, DESTINATION_PORT_RANGE_MIN,
        DESTINATION_PORT_RANGE_MAX, LOGICAL_SOURCE_PORT,
        LOGICAL_DESTINATION_PORT, L7_PARAMETERS,
    ) = (
        'name', 'description', 'protocol', 'ethertype',
        'source_ip_prefix', 'destination_ip_prefix',
        'source_port_range_min', 'source_port_range_max',
        'destination_port_range_min', 'destination_port_range_max',
        'logical_source_port', 'logical_destination_port', 'l7_parameters',
    )

    properties_schema = {
        NAME: properties.Schema(
            properties.Schema.STRING,
            _('Name of the Flow Classifier.'),
            update_allowed=True
        ),
        DESCRIPTION: properties.Schema(
            properties.Schema.STRING,
            _('Description for the Flow Classifier.'),
            update_allowed=True
        ),
        PROTOCOL: properties.Schema(
            properties.Schema.STRING,
            _('IP Protocol for the Flow Classifier.'),
            constraints=[
                constraints.AllowedValues(['tcp', 'udp', 'icmp']),
            ],
        ),
        ETHERTYPE: properties.Schema(
            properties.Schema.STRING,
            _('L2 ethertype.'),
            default='IPv4',
            constraints=[
                constraints.AllowedValues(['IPv4', 'IPv6']),
            ],
        ),
        SOURCE_IP_PREFIX: properties.Schema(
            properties.Schema.STRING,
            _('Source IP prefix or subnet.'),
            constraints=[
                constraints.CustomConstraint('net_cidr')
            ]
        ),
        DESTINATION_IP_PREFIX: properties.Schema(
            properties.Schema.STRING,
            _('Destination IP prefix or subnet.'),
            constraints=[
                constraints.CustomConstraint('net_cidr')
            ]
        ),
        SOURCE_PORT_RANGE_MIN: properties.Schema(
            properties.Schema.INTEGER,
            _('Source protocol port Minimum.'),
            constraints=[
                constraints.Range(1, 65535)
            ]
        ),
        SOURCE_PORT_RANGE_MAX: properties.Schema(
            properties.Schema.INTEGER,
            _('Source protocol port Maximum.'),
            constraints=[
                constraints.Range(1, 65535)
            ]
        ),
        DESTINATION_PORT_RANGE_MIN: properties.Schema(
            properties.Schema.INTEGER,
            _('Destination protocol port minimum.'),
            constraints=[
                constraints.Range(1, 65535)
            ]
        ),
        DESTINATION_PORT_RANGE_MAX: properties.Schema(
            properties.Schema.INTEGER,
            _('Destination protocol port maximum.'),
            constraints=[
                constraints.Range(1, 65535)
            ]
        ),
        LOGICAL_SOURCE_PORT: properties.Schema(
            properties.Schema.STRING,
            _('ID or name of the neutron source port.'),
            constraints=[
                constraints.CustomConstraint('neutron.port')
            ]
        ),
        LOGICAL_DESTINATION_PORT: properties.Schema(
            properties.Schema.STRING,
            _('ID or name of the neutron destination port.'),
            constraints=[
                constraints.CustomConstraint('neutron.port')
            ]
        ),
        L7_PARAMETERS: properties.Schema(
            properties.Schema.MAP,
            _('Dictionary of L7-parameters.'),
            support_status=support.SupportStatus(
                status=support.UNSUPPORTED,
                message=_('Currently, no value is supported for this option.'),
            ),
        ),
    }

    def translation_rules(self, props):
        client_plugin = self.client_plugin()
        return [
            translation.TranslationRule(
                props,
                translation.TranslationRule.RESOLVE,
                [self.LOGICAL_SOURCE_PORT],
                client_plugin=client_plugin,
                finder='find_resourceid_by_name_or_id',
                entity=client_plugin.RES_TYPE_PORT
            ),
            translation.TranslationRule(
                props,
                translation.TranslationRule.RESOLVE,
                [self.LOGICAL_DESTINATION_PORT],
                client_plugin=client_plugin,
                finder='find_resourceid_by_name_or_id',
                entity=client_plugin.RES_TYPE_PORT
            )
        ]

    def _show_resource(self):
        return self.client_plugin().show_ext_resource('flow_classifier',
                                                      self.resource_id)

    def handle_create(self):
        props = self.prepare_properties(
            self.properties,
            self.physical_resource_name())
        flow_classifier = self.client_plugin().create_ext_resource(
            'flow_classifier', props)
        self.resource_id_set(flow_classifier['id'])

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        if prop_diff:
            self.prepare_update_properties(prop_diff)
            self.client_plugin().update_ext_resource('flow_classifier',
                                                     prop_diff,
                                                     self.resource_id)

    def handle_delete(self):
        if self.resource_id is None:
            return
        with self.client_plugin().ignore_not_found:
            self.client_plugin().delete_ext_resource('flow_classifier',
                                                     self.resource_id)


def resource_mapping():
    return {
        'OS::Neutron::FlowClassifier': FlowClassifier,
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import weakref

from keystoneauth1 import exceptions
from keystoneauth1.identity import generic
from keystoneauth1 import plugin
from oslo_config import cfg
from oslo_utils import excutils

import requests

from heat.common import config
from heat.common import exception as heat_exception

cfg.CONF.import_opt('client_retry_limit', 'heat.common.config')


class ClientPlugin(object, metaclass=abc.ABCMeta):

    # Module which contains all exceptions classes which the client
    # may emit
    exceptions_module = None

    # supported service types, service like cinder support multiple service
    # types, so its used in list format
    service_types = []

    # To make the backward compatibility with existing resource plugins
    default_version = None

    supported_versions = []

    def __init__(self, context):
        self._context = weakref.ref(context)
        self._clients = weakref.ref(context.clients)
        self._client_instances = {}
        self._endpoint_existence = {}

    @property
    def context(self):
        ctxt = self._context()
        assert ctxt is not None, "Need a reference to the context"
        return ctxt

    @property
    def clients(self):
        return self._clients()

    _get_client_option = staticmethod(config.get_client_option)

    def client(self, version=None):
        if not version:
            version = self.default_version

        if version in self._client_instances:
            return self._client_instances[version]

        # Back-ward compatibility
        if version is None:
            self._client_instances[version] = self._create()
        else:
            if version not in self.supported_versions:
                raise heat_exception.InvalidServiceVersion(
                    version=version,
                    service=self._get_service_name())

            self._client_instances[version] = self._create(version=version)

        return self._client_instances[version]

    @abc.abstractmethod
    def _create(self, version=None):
        """Return a newly created client."""
        pass

    def _get_region_name(self):
        reg = self.context.region_name or cfg.CONF.region_name_for_services
        # If Shared Services configured, override region for image/volumes
        shared_services_region_name = cfg.CONF.region_name_for_shared_services
        shared_services_types = cfg.CONF.shared_services_types
        if shared_services_region_name:
            if set(self.service_types) & set(shared_services_types):
                reg = shared_services_region_name
        return reg

    def url_for(self, **kwargs):
        keystone_session = self.context.keystone_session

        def get_endpoint():
            return keystone_session.get_endpoint(**kwargs)

        # NOTE(jamielennox): use the session defined by the keystoneclient
        # options as traditionally the token was always retrieved from
        # keystoneclient.
        try:
            kwargs.setdefault('interface', kwargs.pop('endpoint_type'))
        except KeyError:
            pass

        kwargs.setdefault('region_name', self._get_region_name())

        url = None
        try:
            url = get_endpoint()
        except exceptions.EmptyCatalog:
            endpoint = keystone_session.get_endpoint(
                None, interface=plugin.AUTH_INTERFACE)
            token = keystone_session.get_token(None)
            token_obj = generic.Token(endpoint, token)
            auth_ref = token_obj.get_access(keystone_session)
            if auth_ref.has_service_catalog():
                self.context.reload_auth_plugin()
                url = get_endpoint()

        # NOTE(jamielennox): raising exception maintains compatibility with
        # older keystoneclient service catalog searching.
        if url is None:
            raise exceptions.EndpointNotFound()

        return url

    def is_client_exception(self, ex):
        """Returns True if the current exception comes from the client."""
        if self.exceptions_module:
            if isinstance(self.exceptions_module, list):
                for m in self.exceptions_module:
                    if type(ex) in m.__dict__.values():
                        return True
            else:
                return type(ex) in self.exceptions_module.__dict__.values()
        return False

    def is_not_found(self, ex):
        """Returns True if the exception is a not-found."""
        return False

    def is_over_limit(self, ex):
        """Returns True if the exception is an over-limit."""
        return False

    def is_conflict(self, ex):
        """Returns True if the exception is a conflict."""
        return False

    @excutils.exception_filter
    def ignore_not_found(self, ex):
        """Raises the exception unless it is a not-found."""
        return self.is_not_found(ex)

    @excutils.exception_filter
    def ignore_conflict_and_not_found(self, ex):
        """Raises the exception unless it is a conflict or not-found."""
        return self.is_conflict(ex) or self.is_not_found(ex)

    def does_endpoint_exist(self,
                            service_type,
                            service_name):
        endpoint_key = (service_type, service_name)
        if endpoint_key not in self._endpoint_existence:
            endpoint_type = self._get_client_option(service_name,
                                                    'endpoint_type')
            try:
                self.url_for(service_type=service_type,
                             endpoint_type=endpoint_type)
                self._endpoint_existence[endpoint_key] = True
            except exceptions.EndpointNotFound:
                self._endpoint_existence[endpoint_key] = False

        return self._endpoint_existence[endpoint_key]


def retry_if_connection_err(exception):
    return isinstance(exception, requests.ConnectionError)


def retry_if_result_is_false(result):
    return result is False



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from heat.engine import constraints
from heat.engine import properties
from heat.engine.resources.openstack.keystone import project
from heat.engine import stack
from heat.engine import template
from heat.tests import common
from heat.tests import utils

keystone_project_template = {
    'heat_template_version': '2013-05-23',
    'resources': {
        'test_project': {
            'type': 'OS::Keystone::Project',
            'properties': {
                'name': 'test_project_1',
                'description': 'Test project',
                'domain': 'default',
                'enabled': 'True',
                'parent': 'my_father',
                'tags': ['label', 'insignia']
            }
        }
    }
}


class KeystoneProjectTest(common.HeatTestCase):
    def setUp(self):
        super(KeystoneProjectTest, self).setUp()

        self.ctx = utils.dummy_context()

        self.stack = stack.Stack(
            self.ctx, 'test_stack_keystone',
            template.Template(keystone_project_template)
        )

        self.test_project = self.stack['test_project']

        # Mock client
        self.keystoneclient = mock.MagicMock()
        self.test_project.client = mock.MagicMock()
        self.test_project.client.return_value = self.keystoneclient
        self.projects = self.keystoneclient.projects

        # Mock client plugin
        def _id_side_effect(value):
            return value

        keystone_client_plugin = mock.MagicMock()
        keystone_client_plugin.get_domain_id.side_effect = _id_side_effect
        keystone_client_plugin.get_project_id.side_effect = _id_side_effect
        self.test_project.client_plugin = mock.MagicMock()
        self.test_project.client_plugin.return_value = keystone_client_plugin

    def _get_mock_project(self):
        value = mock.MagicMock()
        project_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'
        value.id = project_id
        value.name = 'test_project_1'
        value.domain_id = 'default'
        value.enabled = True
        value.parent_id = 'my_father'
        value.is_domain = False
        return value

    def test_project_handle_create(self):
        mock_project = self._get_mock_project()
        self.projects.create.return_value = mock_project

        # validate the properties
        self.assertEqual(
            'test_project_1',
            self.test_project.properties.get(project.KeystoneProject.NAME))
        self.assertEqual(
            'Test project',
            self.test_project.properties.get(
                project.KeystoneProject.DESCRIPTION))
        self.assertEqual(
            'default',
            self.test_project.properties.get(project.KeystoneProject.DOMAIN))
        self.assertEqual(
            True,
            self.test_project.properties.get(project.KeystoneProject.ENABLED))
        self.assertEqual(
            'my_father',
            self.test_project.properties.get(project.KeystoneProject.PARENT))
        self.assertEqual(
            ['label', 'insignia'],
            self.test_project.properties.get(project.KeystoneProject.TAGS))

        self.test_project.handle_create()

        # validate project creation
        self.projects.create.assert_called_once_with(
            name='test_project_1',
            description='Test project',
            domain='default',
            enabled=True,
            parent='my_father',
            tags=['label', 'insignia'])

        # validate physical resource id
        self.assertEqual(mock_project.id, self.test_project.resource_id)

    def test_properties_title(self):
        property_title_map = {
            project.KeystoneProject.NAME: 'name',
            project.KeystoneProject.DESCRIPTION: 'description',
            project.KeystoneProject.DOMAIN: 'domain',
            project.KeystoneProject.ENABLED: 'enabled',
            project.KeystoneProject.PARENT: 'parent'
        }

        for actual_title, expected_title in property_title_map.items():
            self.assertEqual(
                expected_title,
                actual_title,
                'KeystoneProject PROPERTIES(%s) title modified.' %
                actual_title)

    def test_property_name_validate_schema(self):
        schema = project.KeystoneProject.properties_schema[
            project.KeystoneProject.NAME]
        self.assertEqual(
            True,
            schema.update_allowed,
            'update_allowed for property %s is modified' %
            project.KeystoneProject.NAME)

        self.assertEqual(properties.Schema.STRING,
                         schema.type,
                         'type for property %s is modified' %
                         project.KeystoneProject.NAME)

        self.assertEqual('Name of keystone project.',
                         schema.description,
                         'description for property %s is modified' %
                         project.KeystoneProject.NAME)

    def test_property_description_validate_schema(self):
        schema = project.KeystoneProject.properties_schema[
            project.KeystoneProject.DESCRIPTION]
        self.assertEqual(
            True,
            schema.update_allowed,
            'update_allowed for property %s is modified' %
            project.KeystoneProject.DESCRIPTION)

        self.assertEqual(properties.Schema.STRING,
                         schema.type,
                         'type for property %s is modified' %
                         project.KeystoneProject.DESCRIPTION)

        self.assertEqual('Description of keystone project.',
                         schema.description,
                         'description for property %s is modified' %
                         project.KeystoneProject.DESCRIPTION)

        self.assertEqual(
            '',
            schema.default,
            'default for property %s is modified' %
            project.KeystoneProject.DESCRIPTION)

    def test_property_domain_validate_schema(self):
        schema = project.KeystoneProject.properties_schema[
            project.KeystoneProject.DOMAIN]
        self.assertEqual(
            True,
            schema.update_allowed,
            'update_allowed for property %s is modified' %
            project.KeystoneProject.DOMAIN)

        self.assertEqual(properties.Schema.STRING,
                         schema.type,
                         'type for property %s is modified' %
                         project.KeystoneProject.DOMAIN)

        self.assertEqual('Name or id of keystone domain.',
                         schema.description,
                         'description for property %s is modified' %
                         project.KeystoneProject.DOMAIN)

        self.assertEqual(
            [constraints.CustomConstraint('keystone.domain')],
            schema.constraints,
            'constrains for property %s is modified' %
            project.KeystoneProject.DOMAIN)

        self.assertEqual(
            'default',
            schema.default,
            'default for property %s is modified' %
            project.KeystoneProject.DOMAIN)

    def test_property_enabled_validate_schema(self):
        schema = project.KeystoneProject.properties_schema[
            project.KeystoneProject.ENABLED]
        self.assertEqual(
            True,
            schema.update_allowed,
            'update_allowed for property %s is modified' %
            project.KeystoneProject.DOMAIN)

        self.assertEqual(properties.Schema.BOOLEAN,
                         schema.type,
                         'type for property %s is modified' %
                         project.KeystoneProject.ENABLED)

        self.assertEqual('This project is enabled or disabled.',
                         schema.description,
                         'description for property %s is modified' %
                         project.KeystoneProject.ENABLED)

        self.assertEqual(
            True,
            schema.default,
            'default for property %s is modified' %
            project.KeystoneProject.ENABLED)

    def _get_property_schema_value_default(self, name):
        schema = project.KeystoneProject.properties_schema[name]
        return schema.default

    def test_project_handle_create_default(self):
        values = {
            project.KeystoneProject.NAME: None,
            project.KeystoneProject.DESCRIPTION:
            (self._get_property_schema_value_default(
             project.KeystoneProject.DESCRIPTION)),
            project.KeystoneProject.DOMAIN:
            (self._get_property_schema_value_default(
             project.KeystoneProject.DOMAIN)),
            project.KeystoneProject.ENABLED:
            (self._get_property_schema_value_default(
             project.KeystoneProject.ENABLED)),
            project.KeystoneProject.PARENT:
            (self._get_property_schema_value_default(
             project.KeystoneProject.PARENT)),
            project.KeystoneProject.TAGS:
            (self._get_property_schema_value_default(
             project.KeystoneProject.TAGS))
        }

        def _side_effect(key):
            return values[key]

        mock_project = self._get_mock_project()
        self.projects.create.return_value = mock_project
        self.test_project.properties = mock.MagicMock()
        self.test_project.properties.get.side_effect = _side_effect
        self.test_project.properties.__getitem__.side_effect = _side_effect

        self.test_project.physical_resource_name = mock.MagicMock()
        self.test_project.physical_resource_name.return_value = 'foo'

        # validate the properties
        self.assertEqual(
            None,
            self.test_project.properties.get(project.KeystoneProject.NAME))
        self.assertEqual(
            '',
            self.test_project.properties.get(
                project.KeystoneProject.DESCRIPTION))
        self.assertEqual(
            'default',
            self.test_project.properties.get(project.KeystoneProject.DOMAIN))
        self.assertEqual(
            True,
            self.test_project.properties.get(project.KeystoneProject.ENABLED))
        self.assertIsNone(
            self.test_project.properties.get(project.KeystoneProject.PARENT))

        self.test_project.handle_create()

        # validate project creation
        self.projects.create.assert_called_once_with(
            name='foo',
            description='',
            domain='default',
            enabled=True,
            parent=None,
            tags=[])

    def test_project_handle_update(self):
        self.test_project.resource_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'

        prop_diff = {project.KeystoneProject.NAME: 'test_project_1_updated',
                     project.KeystoneProject.DESCRIPTION:
                     'Test Project updated',
                     project.KeystoneProject.ENABLED: False,
                     project.KeystoneProject.DOMAIN: 'test_domain',
                     project.KeystoneProject.TAGS: ['tag1', 'tag2']}

        self.test_project.handle_update(json_snippet=None,
                                        tmpl_diff=None,
                                        prop_diff=prop_diff)

        self.projects.update.assert_called_once_with(
            project=self.test_project.resource_id,
            name=prop_diff[project.KeystoneProject.NAME],
            description=prop_diff[project.KeystoneProject.DESCRIPTION],
            enabled=prop_diff[project.KeystoneProject.ENABLED],
            domain='test_domain',
            tags=prop_diff[project.KeystoneProject.TAGS]
        )

    def test_project_handle_update_default(self):
        self.test_project.resource_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'

        prop_diff = {project.KeystoneProject.DESCRIPTION:
                     'Test Project updated',
                     project.KeystoneProject.ENABLED: False,
                     project.KeystoneProject.TAGS: ['one', 'two']}

        self.test_project.handle_update(json_snippet=None,
                                        tmpl_diff=None,
                                        prop_diff=prop_diff)

        # validate default name to physical resource name and
        # domain is set from stored properties used during creation.
        self.projects.update.assert_called_once_with(
            project=self.test_project.resource_id,
            name=None,
            description=prop_diff[project.KeystoneProject.DESCRIPTION],
            enabled=prop_diff[project.KeystoneProject.ENABLED],
            domain='default',
            tags=prop_diff[project.KeystoneProject.TAGS]
        )

    def test_project_handle_update_only_enabled(self):
        self.test_project.resource_id = '477e8273-60a7-4c41-b683-fdb0bc7cd151'
        prop_diff = {project.KeystoneProject.ENABLED: False}

        self.test_project.handle_update(json_snippet=None,
                                        tmpl_diff=None,
                                        prop_diff=prop_diff)

        self.projects.update.assert_called_once_with(
            project=self.test_project.resource_id,
            name=None,
            description=None,
            enabled=prop_diff[project.KeystoneProject.ENABLED],
            domain='default',
            tags=['label', 'insignia']
        )

    def test_show_resource(self):
        project = mock.Mock()
        project.to_dict.return_value = {'attr': 'val'}
        self.projects.get.return_value = project
        res = self.test_project._show_resource()
        self.assertEqual({'attr': 'val'}, res)

    def test_get_live_state(self):
        project = mock.Mock()
        project.to_dict.return_value = {
            "is_domain": False,
            "description": "",
            "links": {"self": "link"},
            "enabled": True,
            "id": "8cbb746917ee42f08a787e721552e738",
            "parent_id": "default",
            "domain_id": "default",
            "name": "fake"
        }
        self.projects.get.return_value = project

        reality = self.test_project.get_live_state(
            self.test_project.properties)
        expected = {
            "description": "",
            "enabled": True,
            "domain": "default",
            "name": "fake"
        }
        self.assertEqual(set(expected.keys()), set(reality.keys()))
        for key in expected:
            self.assertEqual(expected[key], reality[key])

    def test_resolve_attributes(self):
        mock_project = self._get_mock_project()
        self.test_project.resource_id = mock_project['id']
        self.projects.get.return_value = mock_project
        self.assertEqual(
            'test_project_1',
            self.test_project._resolve_attribute(
                project.KeystoneProject.NAME_ATTR))
        self.assertEqual(
            'my_father',
            self.test_project._resolve_attribute(
                project.KeystoneProject.PARENT_ATTR))
        self.assertEqual(
            'default',
            self.test_project._resolve_attribute(
                project.KeystoneProject.DOMAIN_ATTR))
        self.assertTrue(self.test_project._resolve_attribute(
            project.KeystoneProject.ENABLED_ATTR))
        self.assertFalse(self.test_project._resolve_attribute(
            project.KeystoneProject.IS_DOMAIN_ATTR))



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from heat.common.i18n import _
from heat.engine.resources.openstack.heat import none_resource
from heat.engine import support


class SaharaJob(none_resource.NoneResource):
    """A resource for creating Sahara Job.

    A job specifies the type of the job and lists all of the individual
    job binary objects. Can be launched using resource-signal.
    """

    support_status = support.SupportStatus(
        version='23.0.0',
        status=support.HIDDEN,
        message=_('Sahara project was retired'),
        previous_status=support.SupportStatus(
            version='22.0.0',
            status=support.DEPRECATED,
            message=_('Sahara project was marked inactive'),
            previous_status=support.SupportStatus(
                version='8.0.0',
                status=support.SUPPORTED
                )))


def resource_mapping():
    return {
        'OS::Sahara::Job': SaharaJob
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common import exception
from heat.common.i18n import _
from heat.engine import constraints
from heat.engine import properties
from heat.engine import resource
from heat.engine import support
from heat.engine import translation


class ClusterTemplate(resource.Resource):
    """A resource for the ClusterTemplate in Magnum.

    ClusterTemplate is an object that stores template information about the
    cluster which is used to create new clusters consistently.
    """

    support_status = support.SupportStatus(version='9.0.0')

    default_client_name = 'magnum'

    entity = 'cluster_templates'

    PROPERTIES = (
        NAME, IMAGE, FLAVOR, MASTER_FLAVOR, KEYPAIR,
        EXTERNAL_NETWORK, FIXED_NETWORK, FIXED_SUBNET, DNS_NAMESERVER,
        DOCKER_VOLUME_SIZE, DOCKER_STORAGE_DRIVER, COE,
        NETWORK_DRIVER, VOLUME_DRIVER, HTTP_PROXY, HTTPS_PROXY,
        NO_PROXY, LABELS, TLS_DISABLED, PUBLIC, REGISTRY_ENABLED,
        SERVER_TYPE, MASTER_LB_ENABLED, FLOATING_IP_ENABLED
    ) = (
        'name', 'image', 'flavor', 'master_flavor', 'keypair',
        'external_network', 'fixed_network', 'fixed_subnet', 'dns_nameserver',
        'docker_volume_size', 'docker_storage_driver', 'coe',
        'network_driver', 'volume_driver', 'http_proxy', 'https_proxy',
        'no_proxy', 'labels', 'tls_disabled', 'public', 'registry_enabled',
        'server_type', 'master_lb_enabled', 'floating_ip_enabled'
    )

    # Change it when magnum supports more function in the future.
    SUPPORTED_VOLUME_DRIVER = {'kubernetes': ['cinder'], 'swarm': ['rexray'],
                               'mesos': ['rexray']}

    properties_schema = {
        NAME: properties.Schema(
            properties.Schema.STRING,
            _('The cluster template name.'),
        ),
        IMAGE: properties.Schema(
            properties.Schema.STRING,
            _('The image name or UUID to use as a base image for cluster.'),
            constraints=[
                constraints.CustomConstraint('glance.image')
            ],
            required=True
        ),
        FLAVOR: properties.Schema(
            properties.Schema.STRING,
            _('The nova flavor name or UUID to use when launching the '
              'cluster.'),
            constraints=[
                constraints.CustomConstraint('nova.flavor')
            ]
        ),
        MASTER_FLAVOR: properties.Schema(
            properties.Schema.STRING,
            _('The nova flavor name or UUID to use when launching the '
              'master node of the cluster.'),
            constraints=[
                constraints.CustomConstraint('nova.flavor')
            ]
        ),
        KEYPAIR: properties.Schema(
            properties.Schema.STRING,
            _('The name of the SSH keypair to load into the '
              'cluster nodes.'),
            constraints=[
                constraints.CustomConstraint('nova.keypair')
            ]
        ),
        EXTERNAL_NETWORK: properties.Schema(
            properties.Schema.STRING,
            _('The external neutron network name or UUID to attach the '
              'Cluster.'),
            constraints=[
                constraints.CustomConstraint('neutron.network')
            ],
            required=True
        ),
        FIXED_NETWORK: properties.Schema(
            properties.Schema.STRING,
            _('The fixed neutron network name or UUID to attach the '
              'Cluster.'),
            constraints=[
                constraints.CustomConstraint('neutron.network')
            ]
        ),
        FIXED_SUBNET: properties.Schema(
            properties.Schema.STRING,
            _('The fixed neutron subnet name or UUID to attach the '
              'Cluster.'),
            constraints=[
                constraints.CustomConstraint('neutron.subnet')
            ]
        ),
        DNS_NAMESERVER: properties.Schema(
            properties.Schema.STRING,
            _('The DNS nameserver address.'),
            constraints=[
                constraints.CustomConstraint('ip_addr')
            ]
        ),
        DOCKER_VOLUME_SIZE: properties.Schema(
            properties.Schema.INTEGER,
            _('The size in GB of the docker volume.'),
            constraints=[
                constraints.Range(min=1),
            ]
        ),
        DOCKER_STORAGE_DRIVER: properties.Schema(
            properties.Schema.STRING,
            _('Select a docker storage driver.'),
            constraints=[
                constraints.AllowedValues(['devicemapper', 'overlay'])
            ],
            default='devicemapper'
        ),
        COE: properties.Schema(
            properties.Schema.STRING,
            _('The Container Orchestration Engine for cluster.'),
            constraints=[
                constraints.AllowedValues(['kubernetes', 'swarm', 'mesos'])
            ],
            required=True
        ),
        NETWORK_DRIVER: properties.Schema(
            properties.Schema.STRING,
            _('The name of the driver used for instantiating '
              'container networks. By default, Magnum will choose the '
              'pre-configured network driver based on COE type.')
        ),
        VOLUME_DRIVER: properties.Schema(
            properties.Schema.STRING,
            _('The volume driver name for instantiating container volume.'),
            constraints=[
                constraints.AllowedValues(['cinder', 'rexray'])
            ]
        ),
        HTTP_PROXY: properties.Schema(
            properties.Schema.STRING,
            _('The http_proxy address to use for nodes in cluster.')
        ),
        HTTPS_PROXY: properties.Schema(
            properties.Schema.STRING,
            _('The https_proxy address to use for nodes in cluster.')
        ),
        NO_PROXY: properties.Schema(
            properties.Schema.STRING,
            _('A comma separated list of addresses for which proxies should '
              'not be used in the cluster.')
        ),
        LABELS: properties.Schema(
            properties.Schema.MAP,
            _('Arbitrary labels in the form of key=value pairs to '
              'associate with cluster.')
        ),
        TLS_DISABLED: properties.Schema(
            properties.Schema.BOOLEAN,
            _('Disable TLS in the cluster.'),
            default=False
        ),
        PUBLIC: properties.Schema(
            properties.Schema.BOOLEAN,
            _('Make the cluster template public. To enable this option, you '
              'must own the right to publish in magnum. Which default set '
              'to admin only.'),
            update_allowed=True,
            default=False
        ),
        REGISTRY_ENABLED: properties.Schema(
            properties.Schema.BOOLEAN,
            _('Enable the docker registry in the cluster.'),
            default=False
        ),
        SERVER_TYPE: properties.Schema(
            properties.Schema.STRING,
            _('Specify the server type to be used.'),
            constraints=[
                constraints.AllowedValues(['vm', 'bm'])
            ],
            default='vm'
        ),
        MASTER_LB_ENABLED: properties.Schema(
            properties.Schema.BOOLEAN,
            _('Indicates whether created clusters should have a load '
              'balancer for master nodes or not.'),
            default=True
        ),
        FLOATING_IP_ENABLED: properties.Schema(
            properties.Schema.BOOLEAN,
            _('Indicates whether created clusters should have a floating '
              'ip or not.'),
            default=True
        ),
    }

    def translation_rules(self, props):
        return [
            translation.TranslationRule(
                props,
                translation.TranslationRule.RESOLVE,
                [self.EXTERNAL_NETWORK],
                client_plugin=self.client_plugin('neutron'),
                finder='find_resourceid_by_name_or_id',
                entity='network'
            ),
            translation.TranslationRule(
                props,
                translation.TranslationRule.RESOLVE,
                [self.FIXED_NETWORK],
                client_plugin=self.client_plugin('neutron'),
                finder='find_resourceid_by_name_or_id',
                entity='network'
            ),
            translation.TranslationRule(
                props,
                translation.TranslationRule.RESOLVE,
                [self.FIXED_SUBNET],
                client_plugin=self.client_plugin('neutron'),
                finder='find_resourceid_by_name_or_id',
                entity='subnet'
            )
        ]

    def validate(self):
        """Validate the provided params."""
        super(ClusterTemplate, self).validate()

        coe = self.properties[self.COE]
        volume_driver = self.properties[self.VOLUME_DRIVER]

        # Confirm that volume driver is supported by Magnum COE per
        # SUPPORTED_VOLUME_DRIVER.
        value = self.SUPPORTED_VOLUME_DRIVER[coe]
        if volume_driver is not None and volume_driver not in value:
            msg = (_('Volume driver type %(driver)s is not supported by '
                     'COE:%(coe)s, expecting a %(supported_volume_driver)s '
                     'volume driver.') % {
                'driver': volume_driver, 'coe': coe,
                'supported_volume_driver': value})
            raise exception.StackValidationFailed(message=msg)

    def handle_create(self):
        args = {
            self.NAME: self.properties[
                self.NAME] or self.physical_resource_name()
        }
        for key in [self.IMAGE, self.FLAVOR, self.MASTER_FLAVOR, self.KEYPAIR,
                    self.EXTERNAL_NETWORK]:
            if self.properties[key] is not None:
                args["%s_id" % key] = self.properties[key]

        for p in [self.FIXED_NETWORK, self.FIXED_SUBNET,
                  self.DNS_NAMESERVER, self.DOCKER_VOLUME_SIZE,
                  self.DOCKER_STORAGE_DRIVER, self.COE, self.NETWORK_DRIVER,
                  self.VOLUME_DRIVER, self.HTTP_PROXY, self.HTTPS_PROXY,
                  self.NO_PROXY, self.LABELS, self.TLS_DISABLED, self.PUBLIC,
                  self.REGISTRY_ENABLED, self.SERVER_TYPE,
                  self.MASTER_LB_ENABLED, self.FLOATING_IP_ENABLED]:
            if self.properties[p] is not None:
                args[p] = self.properties[p]

        ct = self.client().cluster_templates.create(**args)
        self.resource_id_set(ct.uuid)

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        if prop_diff:
            patch = [{'op': 'replace', 'path': '/' + k, 'value': v}
                     for k, v in prop_diff.items()]
            self.client().cluster_templates.update(self.resource_id, patch)
            return self.resource_id

    def check_update_complete(self, id):
        cluster_template = self.client().cluster_templates.get(id)
        if cluster_template.status == 'UPDATE_IN_PROGRESS':
            return False
        elif cluster_template.status == 'UPDATE_COMPLETE':
            return True
        elif cluster_template.status == 'UPDATE_FAILED':
            msg = (_("Failed to update Cluster Template "
                     "'%(name)s' - %(reason)s")
                   % {'name': self.name,
                      'reason': cluster_template.status_reason})
            raise exception.ResourceInError(
                status_reason=msg, resource_status=cluster_template.status)

        else:
            msg = (_("Unknown status updating Cluster Template "
                     "'%(name)s' - %(reason)s")
                   % {'name': self.name,
                      'reason': cluster_template.status_reason})
            raise exception.ResourceUnknownStatus(
                status_reason=msg, resource_status=cluster_template.status)


def resource_mapping():
    return {
        'OS::Magnum::ClusterTemplate': ClusterTemplate
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools

from heat.tests.convergence.framework import reality
from heat.tests.convergence.framework import scenario_template

from oslo_log import log as logging

LOG = logging.getLogger(__name__)


def verify(test, reality, tmpl):
    LOG.info('Verifying %s', tmpl)

    for name in tmpl.resources:
        rsrc_count = len(reality.resources_by_logical_name(name))
        test.assertEqual(1, rsrc_count,
                         'Found %d copies of resource "%s"' % (rsrc_count,
                                                               name))

    all_rsrcs = reality.all_resources()

    for name, defn in tmpl.resources.items():
        phys_rsrc = reality.resources_by_logical_name(name)[0]

        for prop_name, prop_def in defn.properties.items():
            real_value = reality.resource_properties(phys_rsrc, prop_name)

            if isinstance(prop_def, scenario_template.GetAtt):
                targs = reality.resources_by_logical_name(prop_def.target_name)
                prop_def = targs[0].rsrc_prop_data.data[prop_def.attr]
            elif isinstance(prop_def, scenario_template.GetRes):
                targs = reality.resources_by_logical_name(prop_def.target_name)
                prop_def = targs[0].physical_resource_id
            test.assertEqual(prop_def, real_value,
                             'Unexpected value for %s prop %s' % (name,
                                                                  prop_name))

        len_rsrc_prop_data = 0
        if phys_rsrc.rsrc_prop_data:
            len_rsrc_prop_data = len(phys_rsrc.rsrc_prop_data.data)
        test.assertEqual(len(defn.properties),
                         len_rsrc_prop_data)

    test.assertEqual(set(tmpl.resources), set(r.name for r in all_rsrcs))


def scenario_globals(procs, testcase):
    return {
        'test': testcase,
        'reality': reality.reality,
        'verify': functools.partial(verify,
                                    testcase,
                                    reality.reality),

        'Template': scenario_template.Template,
        'RsrcDef': scenario_template.RsrcDef,
        'GetRes': scenario_template.GetRes,
        'GetAtt': scenario_template.GetAtt,

        'engine': procs.engine,
        'worker': procs.worker,
    }
